(*
 * Typechecking for the pi-calculus.
 *)

open Simplify
open Unix

(* this below is to store the name of an agent when checking `create' *)

let thisAgentInfo = ref (Error.unknown : Error.info)
let thisAgentName = ref ("" : string)

let check = Flags.createFlag "check" "*Debug typechecker" false

let verbose = Flags.createFlag "CHECK" "*Debug typechecker verbosely" false

let assertions =
  Flags.createFlag "assertions"
  "*Check all assertions made in the library code" false

let checks =
  Flags.createFlag "checks"
  "*Do all runtime checks specified in the library code" true

let debug = Flags.createFlag "debug"
  "*Include debugging code in libraries" false

let ochErr () =
  Format.print_string "Expected an output-channel type"
let ichErr () =
  Format.print_string "Expected an input-channel type"

let patErr () =
  Format.print_string "pattern's type does not match type of channel"

let ldoutaErr () =
  Format.print_string "expected a Site"

let matchErr () =
  Format.print_string "Expected type does not match actual"

let asErr () =
  Format.print_string
  "Pattern type is not a subtype of the given explicit type"
let atErr () =
  Format.print_string
  "expected a Loc"
let agentErr () =
  Format.print_string "expected an Agent"
let siteErr () =
  Format.print_string "expected a Site"
let timeoutErr () =
  Format.print_string "expected a timeout value of type integer"
let dynamicErr () =
  Format.print_string
  "argument of dynamic does not have required type"
let recpErr () =
  Format.print_string
  "Recursive pattern type does not match explicitly given type"
let recordpErr () =
  Format.print_string
  "Record pattern type does not match annotation"
let outputErr () =
  Format.print_string "value's type does not match type of channel"
let typecaseErr () =
  Format.print_string "argument to typecase is not a Dyn"
let caseErr () =
  Format.print_string "argument to case is not a variant"
let typecasevErr () =
  Format.print_string "branch of typecase does not have expected type"
let casevErr () =
  Format.print_string "branch of case does not have expected type"
let ifErr () =
  Format.print_string "argument to conditional is not a boolean"
let thenElseErr () =
  Format.print_string
  "minimal types of 'then' and 'else' branches must be the same"
let appErr () =
  Format.print_string "argument has incorrect type"
let recordErr () =
  Format.print_string "Expected a record type containing the field "
let valPatErr () =
  Format.print_string "value does not match type of pattern"
let patValErr () =
  Format.print_string "pattern does not match type of value"
let foldErr () =
  Format.print_string "ill-typed fold"
let newErr () =
  Format.print_string "declared type is not a channel type"
let abbrvErr () =
  Format.print_string "type abbreviation fields are not identical"
let boundErr () =
  Format.print_string "bounds do not match on type fields"

exception NeedTypeAnnotation of Error.info

let printReqd = function
  None -> ()
| Some(t,_) -> Format.print_string "required type: "; Types.printT t

let checkReqd fi c t = function
  None -> t
| Some(reqdt,err) -> Leq.isLeq fi c t reqdt err; reqdt

let getReqd fi = function
  None -> Error.errAt fi "Type annotation required"
| Some(t,_) -> t

let getProvidedType fi c topt reqd =
  match (topt,reqd) with
    (None,None) -> raise (NeedTypeAnnotation fi)
  | (Some t,None) -> Kinding.hasKindType c t
  | (None,Some(t,_)) -> t
  | (Some t,Some(treqd,err)) -> 
    let t = Kinding.hasKindType c t in Leq.isLeq fi c treqd t err; t

let rec checkPat' topLevel c reqd = function
  Syntax.RECORDp(fi,l) ->
    (match reqd with
      None ->
        let (l,tl,inc) = synthFieldPat topLevel c 0 l in
        (RECORDp l, Kinding.mkRecord fi tl, inc)
    | Some(reqdt,err) ->
        let tl = Types.getFields fi c reqdt in
        let (l,inc) = checkFieldPat topLevel fi c 0 l tl in
        (RECORDp l, Kinding.mkRecord fi tl, inc))
| Syntax.ASp(fi,n,topt,pat) ->
    let (pat,t,inc) =
      match (reqd,topt) with
        (None,None) -> checkPat topLevel c None pat
      | (_,_) ->
	  let t = getProvidedType fi c topt reqd in
	  checkPat topLevel c (Some(t,asErr)) pat
    in
      if topLevel && Types.isBoundVar c n then
	Error.errAt fi "Attempt to rebind top-level identifier" 
      else
	let x = Var.fresh() in
	(ASp(x,BOXED,pat),t,
	 Types.union (Types.xbind n t (Inter.DYNAMIC x) Types.empty) inc)
| Syntax.VARp(fi,n,topt) ->
    let t = getProvidedType fi c topt reqd in
    if topLevel && Types.isBoundVar c n then
      Error.errAt fi "Attempt to rebind top-level identifier" 
    else
      let x = Var.fresh() in
      (VARp(x,BOXED),t,Types.xbind n t (Inter.DYNAMIC x) Types.empty)
| Syntax.WILDp(fi,topt) ->
    let t = getProvidedType fi c topt reqd in
    (WILDp,t,Types.empty)
| Syntax.RECp(fi,topt,pat) ->
    let t = getProvidedType fi c topt reqd in
    let reqd = Some(Types.unroll fi c t,recpErr) in
    let (pat,_,inc) = checkPat topLevel c reqd pat in
    (pat,t,inc)
| Syntax.ENCODE_PAT(fi,_) -> Error.bugAt fi "Check.checkPat': unexpected ENCODE_PAT"

and synthFieldPat topLevel c i = function
  [] ->
    ([],[],Types.empty)
| Syntax.ANONp(_,pat) :: l ->
    let (pat,t,inc) = checkPat topLevel c None pat in
    let (l,tl,incs) = synthFieldPat topLevel c (i+1) l in
    (ANONp pat :: l, Types.ANON t :: tl,Types.union inc incs)
| Syntax.LABELEDp(_,n,pat) :: l ->
    let (pat,t,inc) = checkPat topLevel c None pat in
    let (l,tl,incs) = synthFieldPat topLevel c (i+1) l in
    (ANONp pat :: l, Types.LABELED(n,t) :: tl,Types.union inc incs)
| Syntax.TYp(fi,n,Syntax.LT t) :: l ->
    let tag = Types.freshTag() in let t = Kinding.wellKinded c t in
    let (l,tl,incs) =
      synthFieldPat topLevel
      (Types.abind (Types.constr c tag (Types.LT t)) n tag) i l
    in
      if topLevel && Types.isBoundTyVar c n then
	Error.errAt fi "Attempt to rebind top-level type identifier" 
      else
	(TYp (Var.fresh()) :: l, Types.TY(n,tag,Types.LT t) :: tl,
	 Types.abind (Types.constr incs tag (Types.LT t)) n tag)
| Syntax.TYp(fi,n,Syntax.EQ t) :: l ->
    let tag = Types.freshTag() and t = Kinding.wellKinded c t in
    let (l,tl,incs) =
      synthFieldPat topLevel
      (Types.abind (Types.constr c tag (Types.EQ t)) n tag) (i+1) l
    in
      if topLevel && Types.isBoundTyVar c n then
	Error.errAt fi "Attempt to rebind top-level type identifier" 
      else
	(TYp (Var.fresh()) :: l, Types.TY(n,tag,Types.EQ t) :: tl,
	 Types.abind (Types.constr incs tag (Types.EQ t)) n tag)

and checkFieldPat topLevel fi c i l tl = match (l,tl) with
  ([],[]) -> ([],Types.empty)
| (Syntax.ANONp(_,pat)::l, Types.ANON(t)::tl) ->
    let (pat,_,inc) = checkPat topLevel c (Some(t,matchErr)) pat in
    let (l,incs) = checkFieldPat topLevel fi c (i+1) l tl in
    (ANONp pat :: l, Types.union inc incs)
| (Syntax.LABELEDp(fi,n,pat)::l, Types.LABELED(reqdn,reqdt)::tl) ->
    if n<>reqdn then Error.errfAt fi (fun () -> 
      Format.print_string "Expected label "; Format.print_string reqdn; 
      Format.print_string " but found "; Format.print_string n);
    let (pat,_,inc) = checkPat topLevel c (Some(reqdt,matchErr)) pat in
    let (l,incs) = checkFieldPat topLevel fi c (i+1) l tl in
    (ANONp pat :: l, Types.union inc incs)
| (Syntax.TYp(fi,n,constr)::l, Types.TY(_,tag,reqdConstr)::tl) ->
    let constr = Kinding.wellKindedConstr c constr in
    Leq.leqConstr fi c reqdConstr constr boundErr;
    let c = Types.abind (Types.constr c tag constr) n tag in
    let (l,incs) = checkFieldPat topLevel fi c i l tl in
    (TYp(Var.fresh()) :: l, Types.abind (Types.constr incs tag constr) n tag)
| (_,_) ->
    Error.errAt fi "Field mismatch in record pattern"

and checkPat topLevel c reqd pat =
  if !check then
    Debug.wrap "checkPat"
    (fun () -> checkPat' topLevel c reqd pat)
    (fun () -> Error.printInfo (Syntax.patInfo pat); printReqd reqd)
    (fun (p,t,inc) -> Types.printT t; if !verbose then Types.printC inc)
  else
    checkPat' topLevel c reqd pat

let safeCheckPat topLevel c reqd pat =
  try checkPat topLevel c reqd pat with NeedTypeAnnotation(fi) ->
  Error.errAt fi "Type annotation required"

let leqReqd fi c topt reqd =
  match (topt,reqd) with
    (None,None) -> None
  | (Some t,None) -> Some(Kinding.hasKindType c t,matchErr)
  | (None,Some(_,_)) -> reqd
  | (Some t,Some(treqd,err)) -> 
    let t = Kinding.hasKindType c t in
    Leq.isLeq fi c t treqd err; Some(t,err)

let rec checkProc' c = function
  Syntax.PRL(_,p,q) ->
    PRL(checkProc c p,checkProc c q)
| Syntax.INPUT(fi,v,pat,p) ->
    let (v,t) = checkVal c None v in
    let argTy = Types.getIChan fi c t in
    let (pat,_,inc) = safeCheckPat false c (Some(argTy,matchErr)) pat in
    INPUT(v,pat,checkProc (Types.union c inc) p)
| Syntax.OUTPUT(fi,v,av) ->
    let (v,t) = checkVal c None v in
    let argTy = Types.getOChan fi c t in
    if Types.leqRChan c t then
      OUTPUT(RESPONSIVE,v,fst(checkVal c (Some(argTy,matchErr)) av))
    else
      OUTPUT(UNKNOWN,v,fst(checkVal c (Some(argTy,matchErr)) av))
| Syntax.INPA(fi,pat,p) ->
    let (v,t) = checkVal c None (Syntax.VARv(fi,!thisAgentName)) in
    let argTy = Types.getIAgent fi c t in
    let (pat,_,inc) = safeCheckPat false c (Some(argTy,matchErr)) pat in
    INPA(v,pat,checkProc (Types.union c inc) p)
| Syntax.OUTA(fi,v,av) ->
    let (v,t) = checkVal c None v in
    let argTy = Types.getOAgent fi c t in
    OUTA(UNKNOWN,v,fst(checkVal c (Some(argTy,matchErr)) av))
| Syntax.LDOUTA(fi,v,sv,av) -> 
    let (v,t) = checkVal c None v in
    let argTy = Types.getOAgent fi c t in
    let (sv,_)= checkVal c (Some(Types.SITE,ldoutaErr)) sv in
    if Types.leqRChan c t then
      LDOUTA(UNKNOWN,v,sv,fst(checkVal c (Some(argTy,matchErr)) av))
    else
      LDOUTA(UNKNOWN,v,sv,fst(checkVal c (Some(argTy,matchErr)) av))
| Syntax.TRYHERE(fi,v,av,p,q) ->
    let (v,t) = checkVal c None v in
    let argTy = Types.getOAgent fi c t in
    if Types.leqRChan c t then
      TRYHERE(v,fst(checkVal c (Some(argTy,matchErr)) av),checkProc c p,checkProc c q)
    else
      TRYHERE(v,fst(checkVal c (Some(argTy,matchErr)) av),checkProc c p,checkProc c q)
| Syntax.REP_INPUT(fi,v,pat,p) ->
    let (v,t) = checkVal c None v in
    let argTy = Types.getIChan fi c t in
    let (pat,_,inc) = safeCheckPat false c (Some(argTy,matchErr)) pat in
    REP_INPUT(v,pat,checkProc (Types.union c inc) p)
| Syntax.TIMED_INPUT(fi,w,v,pat,p,q) ->
    let (w,_) = checkVal c (Some(Types.INT,timeoutErr)) w in
    let (v,t) = checkVal c None v in
    let argTy = Types.getIChan fi c t in
    let (pat,_,inc) = safeCheckPat false c (Some(argTy,matchErr)) pat in
    TIMED_INPUT(w,v,pat,checkProc (Types.union c inc) p,checkProc c q)
| Syntax.REP_TIMED_INPUT(fi,w,v,pat,p,q) ->
    let (w,_) = checkVal c (Some(Types.INT,timeoutErr)) w in
    let (v,t) = checkVal c None v in
    let argTy = Types.getIChan fi c t in
    let (pat,_,inc) = safeCheckPat false c (Some(argTy,matchErr)) pat in
    REP_TIMED_INPUT(w,v,pat,checkProc (Types.union c inc) p,checkProc c q)
| Syntax.LI_SEND(fi,v,chv,av) ->
    let (v,_) = checkVal c (Some(Types.AGENT,agentErr)) v in
    let (chv,cht) = checkVal c None chv in
    let argTy = Types.getOChan fi c cht in
    LI_SEND(v,chv,fst(checkVal c (Some(argTy,matchErr)) av))
| Syntax.REMOTE_SEND(fi,v,sv,chv,av) ->
    let (v,_) = checkVal c (Some(Types.AGENT,agentErr)) v in
    let (sv,_)= checkVal c (Some(Types.SITE,siteErr)) sv in
    let (chv,cht) = checkVal c None chv in
    let argTy = Types.getOChan fi c cht in
    REMOTE_SEND(v,sv,chv,fst(checkVal c (Some(argTy,matchErr)) av))
| Syntax.LOCAL_SEND(fi,v,chv,av) ->
    let (v,_) = checkVal c (Some(Types.AGENT,agentErr)) v in
    let (chv,cht) = checkVal c None chv in
    let argTy = Types.getOChan fi c cht in
    LOCAL_SEND(v,chv,fst(checkVal c (Some(argTy,matchErr)) av))
| Syntax.IFLOCAL(fi,v,chv,av,p,q) ->
    let (v,_) = checkVal c (Some(Types.AGENT,agentErr)) v in
    let (chv,cht) = checkVal c None chv in
    let argTy = Types.getOChan fi c cht in
    IFLOCAL(v,chv,fst(checkVal c (Some(argTy,matchErr)) av),checkProc c p,checkProc c q)
| Syntax.SKIP(_) ->
    SKIP
| Syntax.LET(_,d,p) ->
    let (d,inc) = checkDec false c d in
    LET(d,checkProc (Types.union c inc) p)
| Syntax.IF(_,v,p,q) ->
    let (v,_) = checkVal c (Some(Types.BOOL,ifErr)) v in
    IF(v,checkProc c p,checkProc c q)
| Syntax.TERMINATE(_) ->
    TERMINATE
| Syntax.TYPECASE(_,v,ts,p) ->
    let (v,_) = checkVal c (Some(Types.DYN,typecaseErr)) v in
    let ts = List.map
               (fun (pati,pi) ->
                  let (pati,ti,inci) = safeCheckPat false c (None) pati in
                  let pi = checkProc (Types.union c inci) pi in
                  (pati,ti,pi))
               ts in
    let p = checkProc c p in
    TYPECASE(v,ts,p) 
| Syntax.CASE(_,_,v,ts) ->
(*    let (v,_) = checkVal c (Some(argTy,caseErr)) v in *)
    let (v,_) = checkVal c (None) v in
    let ts = List.map
               (fun (idi,pati,pi) ->
                  let (pati,_,inci) = safeCheckPat false c (None) pati in 
                                          (* Bogus: Some(Types.VART(..),caseErr) instead of None *)
                  let pi = checkProc (Types.union c inci) pi in
                  let x = Var.fresh() in
                  (Var.createVar idi (Var.int_of x),pati,pi))
               ts in
    CASE(v,ts)
| Syntax.ENCODE_PROC(fi,_) -> Error.bugAt fi "Check.checkProc': unexpected ENCODE_PROC"

and checkVal' c reqd = function
  Syntax.CCODEv(fi,ci,l) ->
    let (l,tl) = synthFieldVal c l in
    (match ci.Ccode.kind with
       Ccode.EXP ->
	 (match ci.Ccode.code with
	   ["%async"] ->
	      let async =
		try Sys.getenv "PICTASYNC" = "yes" with Not_found -> false
	      in
		(BOOLv async, checkReqd fi c Types.BOOL reqd)
	 | ["%assertions"] ->
	     (BOOLv (!assertions), checkReqd fi c Types.BOOL reqd)
	 | ["%checks"] ->
	     (BOOLv (!checks), checkReqd fi c Types.BOOL reqd)
	 | ["%debug"] ->
	     (BOOLv (!debug), checkReqd fi c Types.BOOL reqd)
         | _ ->
	     (CCODEv(ci,l), getReqd fi reqd))
     | Ccode.STATEMENT ->
         (CCODEv(ci,l), checkReqd fi c Types.mkUnit reqd)
     | Ccode.NEXP ->
	 (CCODEv(ci,l), getReqd fi reqd)
     | Ccode.NSTATEMENT ->
         (CCODEv(ci,l), checkReqd fi c Types.mkUnit reqd))
| Syntax.RECORDv(fi,l) ->
    (match reqd with
       Some(t,_) ->
         let (_,l) =
	   checkFieldVal fi c l (Types.getFields fi c (Types.clone t)) false
           (fun _ l tl ->
	      match (l,tl) with
	        ([],[]) -> Types.mkUnit
	      | (l, []) ->
		  Error.warningAt fi "Record has extra fields";
		  ignore (synthFieldVal c l); Types.mkUnit
              | ([], _) -> Error.errAt fi "Record does not have enough fields"
	      | (fv :: _,_) -> Error.errAt (Syntax.fvInfo fv)
		"Record field mismatch")
	 in (RECORDv l,t)
     | None ->
	 let (l,tl) = synthFieldVal c l in
	 (RECORDv l, Kinding.mkRecord fi tl))
| Syntax.WITHv(fi,topt,v,l) ->
    let (v,t) =
      match topt with
        None -> checkVal c None v
      | Some t -> checkVal c (Some(Kinding.hasKindType c t,matchErr)) v
    in let fields = Types.getFields fi c t in
    let (l,tl) = synthFieldVal c l in
    let var = Var.fresh() in
    let rec loop i accum =
      if i < 0 then
	LETv(VAL(VARp(var,BOXED),v),RECORDv accum)
      else
	loop (i-1) (ANONv(SELECTv(VARv(Inter.DYNAMIC var),i)) :: accum)
    in
      (loop (List.length fields - 1) l, Kinding.mkRecord fi (fields @ tl))
| Syntax.WHEREv(fi,topt,v,l) ->
    let (v,t) =
      match topt with
        None -> checkVal c None v
      | Some t -> checkVal c (Some(Kinding.hasKindType c t,matchErr)) v
    in let fields = Types.getFields fi c t in
    let (l,tl) = synthFieldVal c l in
    let rcd = Var.fresh() in
    let (d,map) =
      List.fold_left2 (fun (d,map) fv ft ->
	(match ft with
	  Types.LABELED(n,ty) ->
	    let var = Var.fresh() in
	    let (_,i) = Types.fieldType fi c t n in
	    (match fv with
	       ANONv(v) ->
	         (SEQ(d,VAL(VARp(var,BOXED),v)),
		  Misc.IntMap.add i (ANONv(VARv(Inter.DYNAMIC var)),ft) map)
	     | TYv(_) ->
	         Error.bugAt fi "Check.checkVal: WHERE")
	| Types.ANON(_) -> Error.errAt fi "Expected a labeled value field"
	| Types.TY(_) -> Error.errAt fi "Expected a labeled value field")
      ) (VAL(VARp(rcd,BOXED),v),Misc.IntMap.empty) l tl
    in let (_,l,tl) =
      List.fold_right (fun ft (i,l,tl) ->
        let (fv,ft) =
	  try Misc.IntMap.find i map with Not_found ->
	  (ANONv(SELECTv(VARv(Inter.DYNAMIC rcd),i)), ft)
	in
	  (i-1, fv :: l, ft :: tl)
      ) fields (List.length fields-1,[],[])
    in
      (LETv(d,RECORDv l), checkReqd fi c (Types.RECORD tl) reqd)
| Syntax.APPv(fi,topt,v,al) ->
    let (v,t) = checkVal c None v in
    let (res,l) =
      checkFieldVal fi c al
      (Types.getFields fi c (Types.clone (Types.getOChan fi c t))) false
      (fun c l tl ->
	 match (l,tl) with
  	   ([],[Types.ANON t]) ->
	     (match leqReqd fi c topt reqd with
		None -> Types.getOChan fi c t
	      | Some(reqdt,err) ->
		  Leq.isLeq fi c (Types.getOChan fi c t) reqdt err; reqdt)
         | ([],_ :: _ :: _) -> Error.errAt fi "Too few arguments"
         | ([],_ :: _) -> Error.errAt fi "Illegal function type"
         | ([],[]) -> Error.errAt fi "Too many arguments"
         | (fv :: _,[]) -> Error.errAt (Syntax.fvInfo fv) "Too many arguments"
         | (fv :: _,_ :: _) -> Error.errAt (Syntax.fvInfo fv)
	   "Record field mismatch")
    in let
      r = if Types.leqRChan c t then RESPONSIVE else UNKNOWN
    in
      (APPv(r,v,l), res)
| Syntax.SELECTv(fi,v,n) ->
    let (v,t) = checkVal c None v in
    let (ft,i) = Types.fieldType fi c t n in
    (SELECTv(v,i), checkReqd fi c ft reqd)
| Syntax.STRINGv(fi,s) ->
    (STRINGv(s), checkReqd fi c Types.STRING reqd)
| Syntax.SITEv(fi,s,i) ->
    let s = try 
      if (s = "localhost") 
        then string_of_inet_addr((gethostbyname(gethostname())).h_addr_list.(0))
        else string_of_inet_addr((gethostbyname(s)).h_addr_list.(0))
    with Not_found -> Error.errAt fi ("Unknown IP address " ^ s)
    in
    (SITEv(s,i), checkReqd fi c Types.SITE reqd)
| Syntax.CHARv(fi,x) ->
    (CHARv x, checkReqd fi c Types.CHAR reqd)
| Syntax.VARv(fi,x) ->
    let (xt,a) = Types.typeOfX fi x c in
    (VARv a, checkReqd fi c xt reqd)
| Syntax.BOOLv(fi,b) ->
    (BOOLv b, checkReqd fi c Types.BOOL reqd)
| Syntax.INTv(fi,i) ->
    (INTv i, checkReqd fi c Types.INT reqd)
| Syntax.LETv(fi,d,v) ->
    let (d,inc) = checkDec false c d in
    let (v,t) = checkVal (Types.union c inc) reqd v in
    (match reqd with
       None ->
         Error.warningAt fi "Result type for LETv might be wrong";
	 (LETv(d,v),t)
    | Some(reqdt,_) ->
	 (LETv(d,v),reqdt))
| Syntax.RECv(fi,v,topt) ->
    (match leqReqd fi c topt reqd with
       None -> Error.errAt fi "Type annotation required"
     | Some(t,_) -> (fst(checkVal c (Some(Types.unroll fi c t,foldErr)) v),t))
| Syntax.IFv(fi,topt,b,v1,v2) ->
    let (b,_) = checkVal c (Some(Types.BOOL,ifErr)) b in
    let (v1,t) = checkVal c (leqReqd fi c topt reqd) v1 in
    let (v2,_) = checkVal c (Some(t,thenElseErr)) v2 in
    (IFv(b,v1,v2),t)
| Syntax.ABSv(fi,pat,p) ->
    (match reqd with
      None ->
        let (pat,t,inc) = safeCheckPat false c None pat in
        (ABSv(pat,checkProc (Types.union c inc) p),Types.RCHAN t)
    | Some(reqdt,_) ->
        let t = Types.getOChan fi c reqdt in
        let (pat,_,inc) = safeCheckPat false c (Some(t,matchErr)) pat in
        (ABSv(pat,checkProc (Types.union c inc) p),reqdt))
| Syntax.DYNAMICv(fi,v,topt) ->
    let (v,t) =
      match topt with
        None -> checkVal c None v 
      | Some(t) ->
          let t = Kinding.hasKindType c t in
          checkVal c (Some(t,dynamicErr)) v in
    (DYNAMICv(v,t), Types.DYN)
| Syntax.TYPECASEv(fi,topt,v,ts,velse) ->
    let (v,_) = checkVal c (Some(Types.DYN,typecaseErr)) v in
    let (velse,t) = checkVal c (leqReqd fi c topt reqd) velse in
    let ts = List.map
        (fun (pati,vi) ->
          let (pati,ti,inci) = safeCheckPat false c (None) pati in
          let (vi,_) = checkVal (Types.union c inci) 
              (Some(t,typecasevErr)) vi in
          (pati,ti,vi))
        ts in
    (TYPECASEv(v,ts,velse), t)
| Syntax.VARIANTv(fi,n,v) ->
    let (v,t) = checkVal c None v in
    let x = Var.fresh() in
    (VARIANTv(Var.createVar n (Var.int_of x),v), Types.IGNORE) (* Bogus: VART(..) instead of IGNORE? *)
| Syntax.CASEv(fi,topt,v,ts) ->
 (*   let (v,_) = checkVal c (Some(Types.IGNORE,caseErr)) v in *) (* Bogus: VART(..) instead of IGNORE? *)
    let (v,_) = checkVal c None v in  (* Bogus: VART(..) instead of None? *)
    let ts = List.map
        (fun (idi,pati,vi) ->
          let (pati,_,inci) = safeCheckPat false c (None) pati in
          let (vi,_) = checkVal (Types.union c inci) (None) vi in
                                (* Bogus: Some(Types.VART(..),casevErr) instead of None *)
          let x = Var.fresh() in
          (Var.createVar idi (Var.int_of x),pati,vi))
        ts in
    (CASEv(v,ts), Types.IGNORE)   (* Bogus: the actual type of values returned in cases instead of IGNORE *)
| Syntax.ENCODE_VALUE(fi,_) -> Error.bugAt fi "Check.checkVal': unexpected ENCODE_VALUE"

and synthFieldVal c = function
  [] -> ([],[])
| Syntax.ANONv(_,v) :: l ->
    let (v,t) = checkVal c None v in
    let (l,tl) = synthFieldVal c l in
    (ANONv v :: l, Types.ANON t :: tl)
| Syntax.LABELEDv(_,n,v) :: l ->
    let (v,t) = checkVal c None v in
    let (l,tl) = synthFieldVal c l in
    (ANONv v :: l, Types.LABELED(n,t) :: tl)
| Syntax.TYv(fi,t) :: l ->
    let t = Kinding.wellKinded c t in
    let (l,tl) = synthFieldVal c l in
    (TYv BOXED :: l, Types.TY(":x",Types.freshTag(),Types.EQ t) :: tl)

and checkFieldVal fi c l tl synth finalise = match (l,tl) with
  (Syntax.ANONv(_,v) :: l, Types.ANON t :: tl) ->
    if synth then
      let (v,vt) = checkVal c None v in Leq.isLeq fi c vt t matchErr;
      let (res,l) = checkFieldVal fi c l tl true finalise in
      (res, ANONv v :: l)
    else
      let (v,_) = checkVal c (Some(t,matchErr)) v in
      let (res,l) = checkFieldVal fi c l tl false finalise in
      (res, ANONv v :: l)
| (Syntax.LABELEDv(fi,n,v) :: l, Types.LABELED(reqdn,t) :: tl) ->
    if n<>reqdn then Error.errfAt fi (fun () -> 
      Format.print_string "Expected label "; Format.print_string reqdn; 
      Format.print_string " but found "; Format.print_string n);
    if synth then
      let (v,vt) = checkVal c None v in Leq.isLeq fi c vt t matchErr;
      let (res,l) = checkFieldVal fi c l tl true finalise in
      (res, ANONv v :: l)
    else
      let (v,_) = checkVal c (Some(t,matchErr)) v in
      let (res,l) = checkFieldVal fi c l tl false finalise in
      (res, ANONv v :: l)
| (Syntax.TYv(fi,t) :: l, Types.TY(_,tag,constr) :: tl) ->
    let t = Kinding.wellKinded c t in
    Leq.leqConstr fi c (Types.EQ t) constr boundErr;
    let (res,l) =
      checkFieldVal fi (Types.constr c tag (Types.EQ t)) l tl synth finalise
    in
      (Types.subst res tag t, TYv BOXED :: l)
| (_, Types.TY(_,tag,Types.LT t) :: tl) ->
    (match Types.kindOf t with
      Kind.TYPE ->
        let alpha = Types.mkUvar() in
	let (res,l) =
	  checkFieldVal fi (Types.constr c tag (Types.EQ alpha))
	  l tl true finalise
	in
	  Leq.isLeq fi c alpha t boundErr;
	  (Types.subst res tag alpha, TYv BOXED :: l)
    | Kind.ARROW(_,_) ->
        Error.errAt fi "Types arguments of higher kind cannot be omitted")
| (l, Types.TY(_,tag,Types.EQ t) :: tl) ->
    let (res,l) =
      checkFieldVal fi (Types.constr c tag (Types.EQ t)) l tl synth finalise
    in
      (Types.subst res tag t, TYv BOXED :: l)
| (l, tl) ->
    (finalise c l tl, [])

and checkDec' topLevel c = function
  Syntax.SEQ(_,d1,d2) ->
    let (d1,inc1) = checkDec topLevel c d1 in
    let (d2,inc2) = checkDec topLevel (Types.union c inc1) d2 in
    (SEQ(d1,d2), Types.union inc1 inc2)
| Syntax.DEF(_,l) ->
    let (inc,l) = List.fold_left (scanDef topLevel c) (Types.empty,[]) l in
    (DEF(List.map (checkDef (Types.union c inc)) l), inc)
| Syntax.INLINE(fi,n,pat,p) ->
    if topLevel && Types.isBoundVar c n then
      Error.errAt fi "Attempt to rebind top-level identifier" 
    else
      let x = Var.fresh() in
      let (pat,t,inc) = safeCheckPat false c None pat in
      (INLINE(x,pat,checkProc (Types.union c inc) p),
       Types.xbind n (Types.RCHAN t) (Inter.DYNAMIC x) Types.empty)
| Syntax.VAL(_,pat,v) ->
    (try
       let (pat,t,inc) = checkPat topLevel c None pat in
       let (v,_) = checkVal c (Some(t,valPatErr)) v in
       (VAL(pat,v), inc)
     with NeedTypeAnnotation(_) ->
       let (v,t) = checkVal c None v in
       let (pat,_,inc) = safeCheckPat topLevel c (Some(t,patValErr)) pat in
       (VAL(pat,v), inc))
| Syntax.NEW(fi,n,t) ->
    let x = Var.fresh() in
    let t = Kinding.hasKindType c t in
    if not(Types.isChan c t) then
      Error.errfAt fi (fun () ->
       Format.print_string "Expected an channel type but found ";
       Types.displayT t);
    (NEW(x,BOXED), Types.xbind n t (Inter.DYNAMIC x) Types.empty)
| Syntax.RUN(_,p) ->
    (RUN(checkProc c p), Types.empty)
| Syntax.NEWLOC(fi,n) ->
    let x = Var.fresh() in
    (NEWLOC(x), Types.xbind n Types.LOC (Inter.DYNAMIC x) Types.empty)
| Syntax.CREATE(_,l) ->
    let (inc,l) = List.fold_left (scanCreate topLevel c) (Types.empty,[]) l in
    (CREATE(List.map (checkCreate (Types.union c inc)) l), inc)
| Syntax.CREATE_AGENT(_,l) ->
    let (inc,l) = List.fold_left (scanCreateAgent topLevel c) (Types.empty,[]) l in
    (CREATE_AGENT(List.map (checkCreateAgent (Types.union c inc)) l), inc)
| Syntax.AT(fi,x) ->
    let (xt,a) = Types.typeOfX fi x c in
    Leq.isLeq fi c xt Types.LOC atErr; 
    (AT(a), Types.empty)
| Syntax.MIGRATE(fi,v) ->
    let (v,_) = checkVal c (Some(Types.SITE,siteErr)) v in
    (MIGRATE(v), Types.empty)
| Syntax.MACRO(fi,strkey,v) -> Error.bugAt fi "Check.checkDec': unexpected MACRO"
| Syntax.MACROLET(fi,strkey,v,w) -> Error.bugAt fi "Check.checkDec': unexpected MACROLET"
| Syntax.DEFER_DEC(fi,_) -> Error.bugAt fi "Check.checkDec': unexpected DEFER_DEC"
| Syntax.INLINE_DEC(fi,_) -> Error.bugAt fi "Check.checkDec': unexpected INLINE_DEC"
| Syntax.INLINE_VAL(fi,_) -> Error.bugAt fi "Check.checkDec': unexpected INLINE_VAL"
| Syntax.DEF_TYPE(fi,_,_,_) -> Error.bugAt fi "Check.checkDec': unexpected DEF_TYPE"
| Syntax.DEF_DEF(fi,_,_,_,_) -> Error.bugAt fi "Check.checkDec': unexpected DEF_DEF"
| Syntax.DEF_DEC(fi,_,_,_) -> Error.bugAt fi "Check.checkDec': unexpected DEF_DEC"
| Syntax.DEF_PROC(fi,_,_,_) -> Error.bugAt fi "Check.checkDec': unexpected DEF_PROC"
| Syntax.ENCODE_DEC(fi,_) -> Error.bugAt fi "Check.checkDec': unexpected ENCODE_DEC"

and checkDef c (x,pat,inc,p) =
  (x,pat,checkProc (Types.union c inc) p)

and scanDef topLevel c (inc,l) (fi,n,pat,p) =
  if topLevel && Types.isBoundVar c n then
    Error.errAt (Syntax.patInfo pat) "Attempt to rebind top-level identifier" 
  else
    let x = Var.fresh() in
    let (pat,t,patInc) = safeCheckPat false c None pat in
    (Types.xbind n (Types.RCHAN t) (Inter.DYNAMIC x) inc,
     (x,pat,patInc,p) :: l)

and checkCreate c (fi,n, x,p) =
  thisAgentInfo := fi;
  thisAgentName := n;
  (x,checkProc c p)

and scanCreate topLevel c (inc,l) (fi,n,t,p) =
  if topLevel && Types.isBoundVar c n then
    Error.errAt fi "Attempt to rebind top-level identifier" 
  else
    let x = Var.fresh() in
    let t = Kinding.hasKindType c t in
    if not(Types.isAgent c t) then
      Error.errfAt fi (fun () ->
       Format.print_string "Expected an agent type but found ";
       Types.displayT t);
    (Types.xbind n t (Inter.DYNAMIC x) inc,
     (fi,n, x,p) :: l)

and checkCreateAgent c (fi,n, x,p) =
  thisAgentInfo := fi;
  thisAgentName := n;
  (x,checkProc c p)

and scanCreateAgent topLevel c (inc,l) (fi,n,p) =
  if topLevel && Types.isBoundVar c n then
    Error.errAt fi "Attempt to rebind top-level identifier" 
  else
    let x = Var.fresh() in
    (Types.xbind n Types.AGENT (Inter.DYNAMIC x) inc,
     (fi,n, x,p) :: l)

(*
 * Debugging versions of some of the above
 *)  

and checkProc c p =
  if !check then
    Debug.wrap "checkProc"
    (fun () -> checkProc' c p)
    (fun () -> Error.printInfo (Syntax.procInfo p))
    (fun _ -> ())
  else
    checkProc' c p

and checkVal c reqd v = 
  if !check then
    Debug.wrap "checkVal"
    (fun () -> checkVal' c reqd v)
    (fun () -> Error.printInfo (Syntax.valInfo v); printReqd reqd)
    (fun (_,t) -> Types.printT t)
  else
    checkVal' c reqd v

and checkDec topLevel c d =
  if !check then
    Debug.wrap "checkDec"
    (fun cont -> checkDec' topLevel c d)
    (fun () -> Error.printInfo (Syntax.decInfo d))
    (fun (_,inc) -> if !verbose then Types.printC inc)
  else
    checkDec' topLevel c d

let rec addImportProc c = function
  Inter.STRUCT(inc) -> Types.union c inc
| Inter.PRL(p,q) -> addImportProc (addImportProc c p) q
| Inter.VAL(l,p) -> addImportProc (List.fold_left addImportVal c l) p
| Inter.CCODE(_,_,_,p) -> addImportProc c p
| Inter.ATOM(_,_,p) -> addImportProc c p
| Inter.NEW(_,_,p) -> addImportProc c p
| Inter.IF(_) | Inter.INPUT(_) | Inter.OUTPUT(_)
| Inter.SEND(_) | Inter.SKIP -> c
| Inter.ICODE(_,_,p) -> addImportProc c p
| Inter.REP_INPUT(_) | Inter.TIMED_INPUT(_) | Inter.REP_TIMED_INPUT(_) | Inter.IFLOCAL(_) | Inter.TERMINATE -> c
| Inter.MIGRATE(_,_,p) -> addImportProc c p
| _ ->  Error.bug "Error in check.ml: pattern-matching is not exhaustive"
      
and addImportVal c = function
  Inter.DEF(_,_,_,p) -> addImportProc c p
| Inter.TUPLE(_) | Inter.STRING(_) -> c
| Inter.CREATE_AGENT(_,p) -> addImportProc c p

let addImport c prog = addImportProc c prog.Inter.proc

let check imports prog =
  let c = List.fold_left addImport Types.empty imports in
  (* For debugging:
     Format.print_string "Context:"; 
     Types.printC c;
     Format.print_string "Context in O'Caml:";
     Types.caml_formatC c Inter.caml_formatAtom;
  *)
  let (d,inc) = checkDec true c prog.Syntax.dec in
  {includes = prog.Syntax.includes; static = prog.Syntax.static;
   types = inc; dec = d; context = c}
