(*
 * Functions to check for well-kinded type expressions
 * (and to convert tyes into our internal format).
 *)

open Types

let checkDuplicates fi l1 l2 =
  let f seen = function
    LABELED(n,t) ->
      if Misc.StringSet.mem n seen then
	Error.errfAt fi (fun () ->
	  Format.print_string "Duplicate record label: ";
          Format.print_string n
        )
      else
	Misc.StringSet.add n seen
  | ANON(_) | TY(_) ->
      seen
  in let base =
    List.fold_left (fun s ft -> match ft with
        LABELED(n,_) -> Misc.StringSet.add n s
      | ANON(_) | TY(_) -> s
    ) Misc.StringSet.empty l1
  in
    List.fold_left f base l2

let checkVartDuplicates fi l1 l2 =
  let f seen (n,t) =
      if Misc.StringSet.mem n seen then
	Error.errfAt fi (fun () ->
	  Format.print_string "Duplicate variant label: ";
          Format.print_string n
        )
      else
	Misc.StringSet.add n seen
  in let base =
    List.fold_left (fun s ft -> match ft with
        (n,_) -> Misc.StringSet.add n s
    ) Misc.StringSet.empty l1
  in
    List.fold_left f base l2

let rec wellKinded c = function
  Syntax.TVAR(fi,n) ->
    (try
       let tag = Misc.StringMap.find n c.types in
       let t = match Map.find tag c.constr with LT t -> t | EQ t -> t in
       TVAR(Types.kindOf t,n,tag)
     with Not_found ->
       Error.errfAt fi (fun () ->
	 Format.print_string "Unbound type variable: "; Format.print_string n
       ))
| Syntax.TABS(_,l,s) ->
    let (l,kl,c) =
      List.fold_left (fun (l,kl,c) (n,k) ->
        let tag = freshTag() in
	let c = abind (constr c tag (LT(mkTop k))) n tag in
	((n,tag,k) :: l, k :: kl, c)
      ) ([],[],c) l
    in let s = wellKinded c s
    in TABS(Kind.ARROW(List.rev kl,Types.kindOf s),List.rev l,s)
| Syntax.REC(fi,n,k,t) -> 
    let tag = freshTag() in
    let t = wellKinded (abind (constr c tag (LT(mkTop k))) n tag) t in
    if k <> (Types.kindOf t) then
      Error.errfAt fi (fun()->
	Format.print_string "Ill-kinded recursive type:";
	Format.print_space();
	Format.print_string "declared kind of bound variable "; 
	Format.print_string n; 
	Format.print_string " must be equal to kind of corresponding body")
    else
      REC(k,n,tag,t)
| Syntax.TAPP(fi,s,l) ->
    let s = wellKinded c s in let ks = Types.kindOf s in
    let l = List.map (wellKinded c) l in let kl = List.map Types.kindOf l in
    begin
      match ks with
        Kind.ARROW(ks1,ks2) ->
          if kl = ks1 then TAPP(ks2,s,l)
          else Error.errfAt fi (fun()->
            Format.print_string "Ill-kinded type application of type";
            Format.print_space(); printT s; Format.print_string " : ";
	    Kind.print ks; Format.print_space();
	    Format.print_string "to types ";
	    Misc.formatList "" " " "" printT l;
	    Format.print_string " : ";
	    Misc.formatList "" " " "" Kind.print kl)
      | _ -> Error.errfAt fi (fun()->
          Format.print_string "Ill-kinded type application:";
          Format.print_space(); Format.print_string "type ";
          printT s; Format.print_string " has kind "; Kind.print ks;
          Format.print_string ","; Format.print_space();
          Format.print_string "not an arrow kind, so it cannot be applied to";
          Format.print_space(); Misc.formatList "" " " "" printT l)
    end
| Syntax.TOP(_) -> TOP
| Syntax.RECORD(fi,[]) -> RECORD []
| Syntax.RECORD(fi,l) ->
    let (_,l) = List.fold_left buildFieldType (c,[]) l in
    ignore (checkDuplicates fi [] l); RECORD(List.rev l)
| Syntax.WITH(fi,b,l) ->
    let b = wellKinded c b in
    let (_,l) = List.fold_left buildFieldType (c,[]) l in
    (match expose c b with
       RECORD(bl) -> ignore (checkDuplicates fi bl l); RECORD(bl @ List.rev l)
     | _ -> Error.errfAt fi (fun()->
         Format.print_string "Left-hand side of WITH expression"; 
         Format.print_space(); printT b; Format.print_space();
         Format.print_string "is not a record type"))
| Syntax.WHERE(fi,_,_) ->
    Error.bugAt fi "WHERE types not implemented"
| Syntax.RCHAN(fi,t) -> RCHAN(hasKindType c t)
| Syntax.CHAN(fi,t) -> CHAN(hasKindType c t)
| Syntax.OCHAN(fi,t) -> OCHAN(hasKindType c t)
| Syntax.ICHAN(fi,t) -> ICHAN(hasKindType c t)
| Syntax.STRING(_) -> STRING
| Syntax.BOOL(_) -> BOOL
| Syntax.CHAR(_) -> CHAR
| Syntax.INT(_) -> INT
| Syntax.SITE(_) -> SITE
| Syntax.LOC(_) -> LOC
| Syntax.AGENTT(fi,t) -> AGENTT(hasKindType c t)
| Syntax.AGENT(_) -> AGENT
| Syntax.DYN(_) -> DYN
| Syntax.VART(fi,[]) -> VART []
| Syntax.VART(fi,l) ->
    let (_,l) = List.fold_left buildVartFieldType (c,[]) l in
    ignore (checkVartDuplicates fi [] l); VART(List.rev l)
| Syntax.MAP(fi,t1,t2) -> MAP(hasKindType c t1,hasKindType c t2)
| Syntax.ENCODE_TYPE(fi,l) -> ENCODE_TYPE(l)

and buildFieldType (c,l) = function
  Syntax.ANON(_,t) -> (c,ANON(hasKindType c t) :: l)
| Syntax.LABELED(_,n,t) -> (c,LABELED(n,hasKindType c t) :: l)
| Syntax.TY(_,n,con) ->
    let tag = freshTag() and con = wellKindedConstr c con in
    (abind (constr c tag con) n tag,TY(n,tag,con) :: l)

and buildVartFieldType (c,l) (n,t) = (c,(n,hasKindType c t) :: l)

and wellKindedConstr c = function
  Syntax.LT t -> LT(wellKinded c t)
| Syntax.EQ t -> EQ(wellKinded c t)

and hasKindType c tt =
  let t = wellKinded c tt in
  let k = Types.kindOf t in
  if k = Kind.TYPE then t else
  Error.errfAt (Syntax.typeInfo tt) (fun () ->
    Format.print_string "Type "; printT t; Format.print_string " of kind ";
    Kind.print k; Format.print_space();
    Format.print_string "used in a context where only kind Type is allowed")

let mkRecord fi l = ignore (checkDuplicates fi [] l); RECORD(l)


