(*
 *  The runtime system.
 *)

(*

TO DO: 

- DEFs (for ?* at least) out of enviroments (should be faster since less 
  copying, more suitable for any native code generation in the future)

- make sure after "migrate to s P" P is executed first

- make better use of global (which is non-transmissible) and e (transmissible) 
  environments 

- an exception channel to signal, e.g. migration failed (useful in
  the disconnected operations)

*)


open Hashtbl
open Queue
open List

open Unix
(* open ThreadUnix    This is deprecated in OCaml 3.06 *)

open Graphics

open Inter

let log = Flags.createFlag "log" 
   "*print log information" false

let recoveryFile_suffix = ".svp"


(* Types of data structures
   ~~~~~~~~~~~~~~~~~~~~~~~~ *)

type uid = int * string  (* type of globally  unique IDs *)

module AgMap =
  Map.Make(struct
    type t = uid
    let compare = compare
  end)

module AgWait =
  Map.Make(struct
    type t = uid
    let compare = compare
  end)

module ThrMap =
  Map.Make(struct
    type t = int * uid 
    let compare = compare
  end)

module ChanMap =
  Map.Make(struct
    type t = uid
    let compare = compare
  end)

module EnvMap =
  Map.Make(struct
    type t = Var.var
    let compare = compare
  end)

module NameServiceMap =
  Map.Make(struct
    type t = string
    let compare = compare
  end)

module MigLockMap =
  Map.Make(struct
    type t = uid
    let compare = compare
  end)

type agent_t = (uid * heap_t * env_t * runq_t * env_t * proc)  (* transmittable closure *)
and  heap_t  = chanq ChanMap.t          (* store of channel queues *)
and  runq_t  = (env_t * proc) Queue.t   (* run queue of process terms *)
and  env_t   = chanval EnvMap.t         (* environment *)

and chanval = 
    CH of uid                           (* channel ID *)
  | AG of uid                           (* agent ID *)
  | AV of atom                          (* atom value *)
  | CV of value                         (* complex value *)
  | DF of env_t * fundef                (* def value *)
  | LV of chanval list                  (* tuple value *)
  | NIL

and chanq =
    QRS of (env_t * reader) list        (* queue of readers *)
  | QWS of writer list                  (* queue of writers *)

and reader =  
    INP of atom * fundef * status       (* read value from a channel *)
  | REPINP of atom * atom * fundef      (* replicated reader *)
  | TIMER of atom * fundef * int * int  (* timed reader *)

and writer =
    OUT of atom * chanval * status      (* send value to a channel *)

and fundef = value


(* Basic primitives
   ~~~~~~~~~~~~~~~~ *)

let prlog s = 
  if !log then 
    (Format.print_string "[log] "; Format.print_string s; 
     Format.print_newline())
  else ()

let prwarn s =  
  Format.print_string "Warning! "; Format.print_string s; 
  Format.print_newline()

let name (i,n) = (n^"."^(string_of_int i))  (* convert ID to string *)


(*
 * Environment
 *)

(*
let countEnv m =  (* for debugging *)
  let f k v i = i+1 in 
  print_int (EnvMap.fold f m 0); print_newline();
  m
*)

let envCreate = function   
    Some(init) ->
      let f x b e = EnvMap.add x b e in
      EnvMap.fold f init EnvMap.empty
  | None -> 
      EnvMap.empty
	
let envFind x e = try EnvMap.find x e with
  Not_found -> raise (Failure ("envFind: variable "^(Var.name x)^" not bound"))

(**** Removed to permit the environment overwriting which seems necessary 
      to properly deal with environments of nested functions

let envAdd s x b e =
  try 
    let _ = EnvMap.find x e in
    raise (Failure ("envAdd."^s^": variable "^(Var.name x)^" already bound"))
  with Not_found -> EnvMap.add x b e

***)

(* 
 * Heap of channel queues 
 *)

(*
let countChan m =  (* for debugging *)
  let f k v i = i+1 in 
  print_int (ChanMap.fold f m 0); print_newline(); 
  m
*)

let chanCreate = function
    Some(init) ->
      let f x b h = ChanMap.add x b h in
      ChanMap.fold f init ChanMap.empty
  | None -> 
      ChanMap.empty

let chanFind c h = try ChanMap.find c h with
  Not_found -> raise (Failure ("chanFind: channel "^(name c)^" not found"))

let safeChanFind c h = 
  try 
    (h, ChanMap.find c h)
  with 
    Not_found ->
    (* prlog ("safeChanFind: Created a local copy of channel "^(name c)); *)
    (ChanMap.add c (QRS[]) h, (QRS[]))

let chanAdd c q h =
  try 
    let _ = ChanMap.find c h in
    raise (Failure ("chanAdd: channel "^(name c)^" already bound"))
  with Not_found -> ChanMap.add c q h

let chanUpdate c q h = ChanMap.add c q h  (* any previous binding will disappear *)


(* Global data structures
   ~~~~~~~~~~~~~~~~~~~~~~ *)

let sites_l   = ref []                    (* list of sites from a config file *)
let no_sites  = ref 0                     (* total number of sites *)
let site_i    = ref 0                     (* this points to the first free site *)

let serverPortOpt = ref (None : int option) 
let serverPort = ref 0                    (* port number of the I/O server *)

let recoveryFileOpt = ref (None : string option) 
                                          (* recovery file with the runtime system state *)

let trader_hostname = ref None            (* trader's host name *)
let trader_port = ref 0                   (* trader's port number *)

(* not used now
let contbl = Hashtbl.create 255           (* open connections to sites *)
let consem = Mutex.create()               (* semaphore for accessing contbl *)
*)

let npvm_suspended = ref false            (* suspend npvm if agent queue empty *)
let npvm_newagent  = Condition.create()   (* signal if a new agent arrived *)
let npvm_sem       = Mutex.create()       (* semaphore for protecting npvm_newagent *)

let agmap  = ref AgMap.empty              (* store of agents *)
let globe  = ref (envCreate None)         (* global (pervasive) environment *)
let agrunq = Queue.create()               (* run queue of agents *)

(* let agwait = Hashtbl.create 901           (* waiting room for agents with empty process queues *)*)
let agwait = ref AgWait.empty             (* waiting room for agents with empty process queues *)

let agsem  = Mutex.create()               (* semaphore for accessing agmap *)

let funthreads = ref ThrMap.empty         (* system threads executing the non-blocking operations *)
let funsem     = Mutex.create()           (* semaphore for accessing funthreads *)

let tradermap = ref NameServiceMap.empty  (* store of published names *)

let miglockmap = ref MigLockMap.empty     (* locks used in some libs for blocking agent migration *)

let my_host_addr = ref (string_of_inet_addr(
			(gethostbyname(gethostname())).h_addr_list.(0)) : string)

let counter = ref 0
let id_base = ref (((string_of_int (getpid()))^(!my_host_addr)) : string)
let fresh x = incr counter; (!counter,!id_base)  (* create a globally unique ID of a channel queue *)

let counter1 = ref 0
let freshi() = incr counter1; !counter1   (* create a unique IDs of time-outed inputs; 
                                             WARNING: start from 1 since 0 is reserved! *)


(* Evaluation procedures
   ~~~~~~~~~~~~~~~~~~~~~ *)

(* 
 * Evaluate atom to get channel value 
 *)

let rec evalAtom e = function
    INT(a)       -> AV(INT(a))   (* return Integer *)
  | CHAR(a)      -> AV(CHAR(a))  (* return Char *)
  | BOOL(a)      -> AV(BOOL(a))  (* return Boolean *)
  | STATIC(x)    -> getChanVal e x
  | DYNAMIC(x)   -> getChanVal e x
  | PROJECT(a,i) -> 
      begin
	try match evalAtom e a with
	  LV(l)  -> evalChanVal e (List.nth l i) 
	| CV(v)  ->
	    begin
	      match v with
		TUPLE(_,l) -> evalAtom e (List.nth l i)
	      | STRING(_) | DEF(_)  | CREATE_AGENT(_) ->
		  raise (Failure "evalAtom: expected tuple")
	    end
        | DF(_)  -> raise (Failure "evalAtom: expected complex value but found DF")
	| AV(_)  -> raise (Failure "evalAtom: expected complex value but found atom")
	| CH(_)  -> raise (Failure "evalAtom: expected complex value but found CH")
	| AG(_)  -> raise (Failure "evalAtom: expected complex value but found AG")
	| NIL    -> raise (Failure "evalAtom: nil binding")
	with Not_found -> raise (Failure "evalAtom: failed in PROJECT")
      end
  | ADDR(_) -> raise (Failure "evalAtom: found ADDR")
  | CCONST(_) -> raise (Failure "evalAtom: found CCONST")

(* 
 * Evaluate channel value 
 *)
	
and evalChanVal e = function
    CH(c) -> CH(c)  (* return channel ID *)
  | AG(n) -> AG(n)  (* return agent ID *)
  | AV(a) -> AV(a)  (* return atom value *)
  | LV(l) -> LV(l)  (* return tuple value *)
  | CV(v) ->
      begin
	match v with
      	| STRING(x,s) -> CV(STRING(x,s))              (* return string *)
      	| TUPLE(x,l)  -> LV(List.map (evalAtom e) l)  (* return tuple value *)
      	| DEF(_) | CREATE_AGENT(_) -> 
	    raise (Failure "evalChanVal: expected string or tuple")
      end
  | DF(e,f) -> DF(e,f)  (* return (env,def) value *)
  | NIL -> raise (Failure "evalChanVal: nil binding")

(* 
 * Get channel value from environment 
 *)

and getChanVal (e,le) x =
  try evalChanVal (e,le) (EnvMap.find x le)

  with Not_found ->
    try evalChanVal (e,le) (EnvMap.find x e)
    with Not_found -> 
      try evalChanVal (e,le) (EnvMap.find x !globe)
      with Not_found -> raise (Failure ("getChanVal: variable "^(Var.name x)^" not found"))

(* 
 * Extract channel ID from channel value 
 *)

let getCH = function
    CH(c) -> c  (* return channel ID *)
  | AG(n) -> print_string (name n); raise (Failure "getCH: expected channel ID but found agent ID")
  | AV(a) -> caml_formatAtom a; raise (Failure "getCH: expected channel ID but found atom value")
  | CV(v) -> caml_formatVal (fun _ -> ()) v; raise (Failure "getCH: expected channel ID but found complex value")
  | DF(e,d) -> raise (Failure "getCH: expected channel ID but found (env_t,def)")
  | LV(_) -> raise (Failure "getCH: expected channel ID but found tuple value")
  | NIL   -> raise (Failure "getCH: nil binding")

(* 
 * Extract agent ID from channel value 
 *)

let getAG = function
    AG(n) -> n  (* return agent ID *)
  | CH(c) -> print_string (name c); raise (Failure "getAG: expected agent ID but found channel ID")
  | AV(a) -> caml_formatAtom a; raise (Failure "getAG: expected agent ID but found atom value")
  | CV(v) -> caml_formatVal (fun _ -> ()) v; raise (Failure "getAG: expected agent ID but found complex value")
  | DF(e,d) -> raise (Failure "getAG: expected agent ID but found (env_t,def)")
  | LV(_) -> raise (Failure "getAG: expected agent ID but found a tuple value")
  | NIL   -> raise (Failure "getAG: nil binding")

(* 
 * Extract (e,def) from channel value 
 *)

let getDF = function
    DF(e,d) -> (e,d)
  | CH(c) -> print_string (name c); raise (Failure "getDF: expected (env_t,def) but found channel ID")
  | AG(n) -> print_string (name n); raise (Failure "getDF: expected (env_t,def) but found agent ID")
  | AV(a) -> caml_formatAtom a; raise (Failure "getDF: expected (env_t,def) but found atom value")
  | CV(v) -> caml_formatVal (fun _ -> ()) v; raise (Failure "getDF: expected (env_t,def) but found complex value")
  | LV(_) -> raise (Failure "getDF: expected (env_t,def) but found tuple value")
  | NIL   -> raise (Failure "getDF: nil binding")

(* 
 * Extract Internet name and port number of the site from channel value 
 *)

let getSite = function 
    LV(l) ->  (* return a tuple of Internet name and port number *)
      (match l with
      	(CV(STRING(_,s))::AV(INT(i))::[]) -> (s,i)
      |	_ -> raise (Failure "getSite: Internet name & port number format not recognised"))
  | CH(c) -> print_string (name c); raise (Failure "getSite: expected Internet name & port number but found channel ID")
  | AG(n) -> print_string (name n); raise (Failure "getSite: expected Internet name & port umber but found agent ID")
  | AV(a) -> caml_formatAtom a; raise (Failure "getSite: expected Internet name & port number but found atom value")
  | CV(v) -> caml_formatVal (fun _ -> ()) v; raise (Failure "getSite: expected Internet name & port number but found complex value")
  | DF(e,d) -> raise (Failure "getSite: expected Internet name & port number but found complex value")
  | NIL   -> raise (Failure "getSite: nil binding")


(*
 * Utilities
 *)

let boolof e a =
  match evalAtom e a with
  | AV(BOOL(b)) -> b  (* return Boolean *)
  | AV(_) | CH(_) | AG(_) | CV(_) | DF(_) | LV(_) -> raise (Failure "boolof: expected Boolean")
  | NIL -> raise (Failure "boolof: nil binding")

let intof e a =
  match evalAtom e a with
  | AV(INT(i)) -> i  (* return integer *)
  | AV(_) | CH(_) | AG(_) | CV(_) | DF(_) | LV(_) -> raise (Failure "intof: expected integer")
  | NIL -> raise (Failure "intof: nil binding")

let charof e a =
  match evalAtom e a with
  | AV(CHAR(ch)) -> ch  (* return character *)
  | AV(_) | CH(_) | AG(_) | CV(_) | DF(_) | LV(_) -> raise (Failure "charof: expected character")
  | NIL -> raise (Failure "charof: nil binding")

let stringof e a =
  match evalAtom e a with
  | CV(STRING(x,s)) -> s  (* return string *)
  | CV(_) | DF(_) | CH(_) | AG(_) | AV(_) | LV(_) -> raise (Failure "stringof: expected string")
  | NIL -> raise (Failure "stringof: nil binding")

let listof e a =
  match evalAtom e a with
  | LV(l) -> l  (* return tuple *)
  | CH(_) | AG(_) | AV(_) | CV(_) | DF(_) -> raise (Failure "listof: expected tuple")
  | NIL -> raise (Failure "listof: nil binding")

let atomof e a =
  match evalAtom e a with
  | AV(a) -> a  (* return atom *)
  | CH(_) | AG(_) | CV(_) | DF(_) | LV(_) -> raise (Failure "atomof: expected atom")
  | NIL -> raise (Failure "atomof: nil binding")

let chanof e a =
  match evalAtom e a with
  | CH(c) -> c  (* return channel ID *)
  | AV(_) | AG(_) | CV(_) | DF(_) | LV(_) -> raise (Failure "atomof: expected atom")
  | NIL -> raise (Failure "atomof: nil binding")

let variantof e a = 
  match evalChanVal e a with
    LV(CV(STRING(_,s))::a::[]) -> (s,a)  (* return label and value *)
  | _ -> raise (Failure "variantof: expected variant")

let nilof e a = 
  match evalAtom e a with
    NIL -> true
  | AV(_) | CH(_) | AG(_) | CV(_) | DF(_) | LV(_) -> false



(**********************
 *  System libraries  *
 **********************)


let update_env (e,le) result ropt = 
  match ropt with
    (Some rc) -> 
        if result <> NIL then  (* store result in a local environment *)
(*
          (e,envAdd "update_env" rc result le)
*)
          (e,EnvMap.add rc result le)
        else
          raise (Failure ("update_env: expected a result value"))
    | None -> 
        if result <> NIL then
          raise (Failure ("update_env: name of a result channel not specified"))
        else (e,le)


let evalCcode (ag,e,ropt,ci,args,hrp) =

  let get_core result =
    let (e,le) = e in
    let (e,le) = update_env (e,le) result ropt in
    let (h,r,p) = hrp in
    Mutex.lock agsem;
    let agr = Queue.create() in
    Queue.add (le,p) agr;
    Queue.iter (fun x -> Queue.add x agr) r;
    let map = AgMap.add ag (h,e,agr) !agmap in
    let runq = Queue.create() in
    Queue.add ag runq;
    Queue.iter (fun x -> Queue.add x runq) agrunq;
    let core = (map,runq,!agwait,!tradermap) in
    Mutex.unlock agsem;
    core
  in      
  
  let load core =   (* Descruct all current agents and load new agents *)
    let kill (tid,a) (t,state,cont) = 
      try
	Thread.kill t;
        prerr_endline "Nomadic Pict: I/O thread killed (since agent vanished)."
      with
	_ -> 
          prerr_endline "Nomadic Pict: Kill thread."
    in
    let (agmap_f,agrunq_f,agwait_f,tradermap_f) = core in
    Mutex.lock agsem;
    Mutex.lock funsem; 
    ThrMap.iter kill !funthreads;  (* kill any system threads of current agents *)
    funthreads := ThrMap.empty;
    Mutex.unlock funsem;
    agmap := AgMap.empty;
    agmap := AgMap.map (fun a -> a) agmap_f;      
    Queue.clear agrunq;
    Queue.transfer agrunq_f agrunq;
    agwait := AgWait.empty;
    agwait := AgWait.map (fun a -> a) agwait_f;
    tradermap := NameServiceMap.map (fun a -> a) tradermap_f;
    Mutex.unlock agsem
  in
  
  (* Standard library functions
     ~~~~~~~~~~~~~~~~~~~~~~~~~~ *)
  
let libBool = function
    ("&&",x::y::[]) -> AV (BOOL((boolof e x) && (boolof e y)))
  | ("||",x::y::[]) -> AV (BOOL((boolof e x) || (boolof e y)))
  | ("xor",x::y::[]) ->
      let a = boolof e x
      and b = boolof e y 
      in let c = ((not a) && b) || (a && (not b)) in
      AV (BOOL c)
  | ("not",x::[]) -> AV (BOOL(not (boolof e x)))
  | (s,_) -> raise (Failure ("Function bool."^s^" not defined or incorrect number of arguments"))

and libChar = function
    ("getChar",c::[]) -> AV (CHAR(Char.chr (intof e c)))
  | ("toString",c::[]) -> CV (STRING(Var.fresh(),Char.escaped (charof e c)))
  | ("hash",c::[]) -> AV (INT(Char.code (charof e c)))
  | ("pr",c::[]) -> output_char Stdlib.stdout (charof e c); NIL
  | (s,_) -> raise (Failure ("Function char."^s^" not defined or incorrect number of arguments"))

and libError = function
(*    ("exit",n::[]) -> exit (intof e n); NIL *)
    ("exit",n::[]) -> exit (intof e n)
  | ("fatal",s::[]) -> output_string Stdlib.stderr ("Fatal error: "^(stringof e s));
(*      exit 1; NIL *)
      exit 1
  | (s,_) -> raise (Failure ("Error: "^s^" not defined or incorrect number of arguments"))

and libInt = function
    ("-", x::y::[]) -> AV (INT((intof e x) - (intof e y)))
  | ("*", x::y::[]) -> AV (INT((intof e x) * (intof e y)))
  | ("div", x::y::[]) -> AV (INT((intof e x) / (intof e y)))
  | ("mod", x::y::[]) -> AV (INT((intof e x) mod (intof e y)))
  | ("dec", x::[]) -> AV (INT(pred(intof e x)))
  | ("inc", x::[]) -> AV (INT(succ(intof e x)))
  | ("sgn", x::[]) -> 
      let x = intof e x in
      if (x = 0) then (AV (INT(0)))
      else 
	(if (x < 0) then (AV (INT(-1)))
	else (AV (INT(1))))
  | ("abs", x::[]) -> AV (INT(abs(intof e x)))
  | ("land", x::y::[]) -> AV (INT((intof e x) - (intof e y)))
  | ("lor", x::y::[]) -> AV (INT((intof e x) * (intof e y)))
  | ("lxor", x::y::[]) -> AV (INT((intof e x) - (intof e y)))
  | ("lnot", x::y::[]) -> AV (INT((intof e x) * (intof e y)))
(*
  | ("bitIsSet", x::y::[]) -> AV (INT((intof e x) - (intof e y)))
*)
  | ("max", x::y::[]) -> AV (INT((intof e x) * (intof e y)))
  | ("min", x::y::[]) -> AV (INT((intof e x) - (intof e y)))
  | ("fromString",x::[]) -> AV (INT(int_of_string (stringof e x)))
  | ("toString",x::[]) -> CV (STRING(Var.fresh(), string_of_int (intof e x)))
  | ("pr",x::[]) -> print_int (intof e x); NIL
  | (s,_) -> raise (Failure ("Function int."^s^" not defined or incorrect number of arguments"))

and libList = function
  | ("nil",[]) -> LV ([])
  | ("cons",hd::tl::[]) -> 
      let h = evalAtom e hd
      and t = listof e tl in LV (h::t)
  | ("null",al::[]) -> 
      let al = listof e al in
      if (al = []) then AV (BOOL(true)) else AV (BOOL(false))
  | ("getCar",al::[]) ->
      let al = listof e al in (List.hd al)
  | ("getCdr",al::[]) ->
      let al = listof e al in LV (List.tl al)
  | (s,_) -> raise (Failure ("Function list."^s^" not defined or incorrect number of arguments"))

and libMisc = function
    ("pr", x::[]) -> output_string Stdlib.stdout (stringof e x); NIL
  | ("prNL", x::[]) -> output_string Stdlib.stdout (stringof e x); output_string Stdlib.stdout "\n"; NIL
  | ("prErr", x::[]) -> output_string Stdlib.stderr (stringof e x); NIL
  | ("prErrNL", x::[]) -> output_string Stdlib.stderr (stringof e x); output_string Stdlib.stderr "\n"; NIL
  | ("flush", []) -> flush Stdlib.stdout; NIL
  | ("printi", x::[]) -> Printf.printf "%d\n" (intof e x); flush Stdlib.stdout; NIL
  | (s,_) -> raise (Failure ("Function misc."^s^" not defined or incorrect number of arguments"))

and libOpt = function 
    ("no", []) -> AV (BOOL false)
  | ("null", o::[]) -> if nilof e o then AV(BOOL true) else AV(BOOL false)
  | (s,_) -> raise (Failure ("Function opt."^s^" not defined or incorrect number of arguments"))

and libPrim = function
    ("+", x::y::[])  ->  AV (INT((intof e x) + (intof e y)))
  | ("==", x::y::[]) -> (AV (BOOL((intof e x) =  (intof e y))))
  | ("<>", x::y::[]) -> (AV (BOOL((intof e x) <> (intof e y))))
  | (">>", x::y::[]) -> (AV (BOOL((intof e x) >  (intof e y))))
  | ("<<", x::y::[]) -> (AV (BOOL((intof e x) <  (intof e y))))
  | (">=", x::y::[]) -> (AV (BOOL((intof e x) >= (intof e y))))
  | ("<=", x::y::[]) -> (AV (BOOL((intof e x) <= (intof e y))))
  | ("size", x::[]) -> (AV (INT (String.length (stringof e x))))
  | ("+$", x::y::[]) -> 
      begin match ropt with
	(Some rc) -> (CV (STRING(rc,String.concat "" ((stringof e x)::(stringof e y)::[]))))
      |	None -> raise (Failure "prim.+$: name of result channel not specified")
      end
  | ("==$", x::y::[]) -> (AV (BOOL((stringof e x) = (stringof e y))))
  | (s,_) -> raise (Failure ("Function prim."^s^" not defined or incorrect number of arguments"))

and libString = function
    ("setNth",s::n::c::[]) -> 
      let s = stringof e s
      and n = intof e n
      and c = charof e c in
(*      s.[n] <- c; NIL  to samo co String.set s n c *)
      Bytes.set (Bytes.of_string s) n c; NIL
  | ("make",n::c::[]) -> (CV (STRING(Var.fresh(),String.make (intof e n) (charof e c))))
  | ("createf",n::[]) -> (CV (STRING(Var.fresh(),Bytes.to_string (Bytes.create (intof e n)))))
  | ("nth",s::n::[]) -> (AV (CHAR(String.get (stringof e s) (intof e n))))
  | ("sub",s::start::len::[]) -> (CV (STRING(Var.fresh(),String.sub (stringof e s) 
						  (intof e start) (intof e len))))
  | ("cmp",s1::s2::[]) -> (AV (INT(compare (stringof e s1) (stringof e s2))))
      (*
      	 let b = (s1 = s2) in
      	 if b then (AV (INT(0)))
      	 else let b = (s1 < s2) in
      	 if b then (AV (INT(-1))) else (AV (INT(1))) 
      *)
  | ("<>$",s1::s2::[]) -> AV (BOOL((stringof e s1) <> (stringof e s2)))
  | (">>$",s1::s2::[]) -> AV (BOOL((stringof e s1)  > (stringof e s2)))
  | ("<<$",s1::s2::[]) -> AV (BOOL((stringof e s1)  < (stringof e s2)))
  | (">=$",s1::s2::[]) -> AV (BOOL((stringof e s1) >= (stringof e s2)))
  | ("<=$",s1::s2::[]) -> AV (BOOL((stringof e s1) <= (stringof e s2)))
  | ("occursIn",s::c::[]) -> 
      let s = stringof e s in
      let c = charof e c in
      let n = String.length s in
      let rec loop i = 
	if (i=n) then
	  false
	else
	  if (s.[i] = c) 
      	  then true 
          else loop (i+1)
      in
      AV (BOOL(loop 0))
  | ("toInt",s::[]) -> 
      (try
	LV (AV(BOOL(true))::AV(INT(int_of_string (stringof e s)))::[])
      with _ ->
	LV (AV(BOOL(false))::AV(INT(0))::[]))
  | (s,_) -> raise (Failure ("Function string."^s^" not defined or incorrect number of arguments"))

and libSys = function
    ("==",a::b::[]) ->
      let a = evalAtom e a 
      and b = evalAtom e b
      in if (a = b) then (AV (BOOL(true))) else (AV (BOOL(false)))
  | ("eqs",a::b::[]) ->
      let a = getSite (evalAtom e a)
      and b = getSite (evalAtom e b)
      in if (a = b) then (AV (BOOL(true))) else (AV (BOOL(false)))
  | ("<>",a::b::[]) ->
      let a = evalAtom e a 
      and b = evalAtom e b
      in if (a <> b) then (AV (BOOL(true))) else (AV (BOOL(false)))
  | ("chan_name",c::[]) -> 
      CV(STRING(Var.fresh(),(name (chanof e c))))
  | ("addr_of_site",s::[]) -> 
      let (inet_name,_) = getSite (evalAtom e s) in 
      CV(STRING(Var.fresh(),inet_name))
  | ("port_of_site",s::[]) -> 
      let (_,port) = getSite (evalAtom e s) in 
      AV(INT(port))
  | ("get_next_site",[]) ->
      begin
        if !no_sites = 0 then
          raise (Failure ("sys.get_next_site: no sites in a config file"))
        else  
          let (inet_name,port) = List.nth !sites_l !site_i in
          incr site_i;
          if !site_i = !no_sites then site_i := 0;
	  LV (CV(STRING(Var.fresh(),inet_name))::AV(INT(port))::[])
      end
  | ("get_site",i::[]) ->
      begin
	try 
          let (inet_name,port) = List.nth !sites_l (intof e i) in
	  LV (CV(STRING(Var.fresh(),inet_name))::AV(INT(port))::[])
	with
	  _ -> raise (Failure ("No config file or out of range."))
      end
  | ("gethostname",[]) -> CV(STRING(Var.fresh(),gethostname()))
  | ("this_site",[]) -> LV(CV(STRING(Var.fresh(),!my_host_addr))::AV(INT(!serverPort))::[])
  | ("this_agent",[]) -> AG(ag)
  | ("agent_id",[]) -> CV(STRING(Var.fresh(),name ag))
      (* let id = getAG (evalAtom e a) in *)
  | ("is_here",a::[]) ->
      let a = getAG (evalAtom e a) in
      (try
	Mutex.lock agsem;
	(* AgMap.find a !agmap *)
    	let _ = AgMap.find a !agmap in (); (* to avoid warning *)
    	Mutex.unlock agsem;
	AV (BOOL true)
      with
    	Not_found -> (* agent is not here *)
	  Mutex.unlock agsem;
	  AV (BOOL false))
  | ("time",[]) -> AV (INT (Stdlib.truncate (time())))
  | ("sleep",t::[]) -> sleep (intof e t); NIL
  | ("saveall",p::[]) -> 
      let prefix = (stringof e p) in
      let (filename,oc) = (Filename.open_temp_file ~mode:[Open_binary] prefix recoveryFile_suffix) in
      let result = CV(STRING(Var.fresh(),(Filename.chop_suffix filename recoveryFile_suffix))) in
      output_value oc (get_core result);
      close_out oc;
      result
  | ("saveall_unsafe",s::[]) -> 
      let filename = ((stringof e s)^recoveryFile_suffix) in
      (try
	let oc = (open_out_bin filename) in
	output_value oc (get_core NIL);
	close_out oc;
      with
	Sys_error(err) -> Error.err ("Couldn't open " ^ filename ^ " for output " ^ "(" ^ err ^ ")")
      );
      NIL
  | ("load",fn::[]) -> 
      let filename = ((stringof e fn)^recoveryFile_suffix) in
      let ic = 
	try
	  (open_in_bin filename) 
	with
	  Sys_error(err) -> Error.err ("Couldn't open " ^ filename ^ " for input " ^ "(" ^ err ^ ")")
      in
      let core = input_value ic in
      let _ = load core in
      NIL
  | ("recover",[]) -> 
      (match !recoveryFileOpt with
	Some s -> 
	  let filename = (s^recoveryFile_suffix) in
	  let ic = 
	    try
	      (open_in_bin filename)
	    with
	      Sys_error(err) -> Error.err ("Couldn't open " ^ filename ^ " for input " ^ "(" ^ err ^ ")")
	  in
	  let core = input_value ic in
	  let _ = load core in
	  NIL
      | None -> NIL)
  | ("read_line",[]) -> CV(STRING(Var.fresh(),read_line()))
  | ("read_int",[]) -> AV (INT (read_int()))
  | ("lookup",s::[]) ->
      (try 
	let chv = NameServiceMap.find (stringof e s) !tradermap in
  	LV (AV(BOOL(true))::chv::[])
      with Not_found -> 
	LV (AV(BOOL(false))::NIL::[]))
  | ("publish",key::v::[]) -> 
      tradermap := NameServiceMap.add (stringof e key) (evalAtom e v) !tradermap; NIL
  | ("name_service",[]) ->
      (match !trader_hostname with
	Some s -> 
	  LV (AV(BOOL(false))::
	      LV (CV(STRING(Var.fresh(),s))::AV(INT(!trader_port))::[])::
	      [])
      | None ->
	  LV (AV(BOOL(true))::
	      LV (CV(STRING(Var.fresh(),Unix.gethostname()))::AV(INT(!serverPort))::[])::
	      []))
  | ("block_migration",[]) -> miglockmap := MigLockMap.add ag 0 !miglockmap; NIL
  | ("release_migration",[]) -> miglockmap := MigLockMap.remove ag !miglockmap; NIL
  | (s,_) -> raise (Failure ("Function sys."^s^" not defined or incorrect number of arguments"))


(* O'Caml Graphics 
   ~~~~~~~~~~~~~~~ *)

and libGraphics = function
    ("open_graph",name::[]) -> 
      (try
      	open_graph (stringof e name);
 	LV (AV(BOOL(true))::CV(STRING(Var.fresh(),""))::[])
      with Graphics.Graphic_failure(s) ->
 	LV (AV(BOOL(false))::CV(STRING(Var.fresh(),s))::[]))
  | ("close_graph",[]) -> close_graph(); NIL
  | ("clear_graph",[]) -> clear_graph(); NIL
  | ("size_x",[]) -> AV(INT(size_x()))
  | ("size_y",[]) -> AV(INT(size_y()))
  | ("rgb",r::g::b::[]) -> AV(INT(rgb (intof e r) (intof e g) (intof e b)))
  | ("set_color",c::[]) -> set_color (intof e c); NIL
  | ("plot",x::y::[]) -> plot (intof e x)(intof e y); NIL
  | ("point_color",x::y::[]) -> AV(INT(point_color (intof e x) (intof e y)))
  | ("moveto",x::y::[]) -> moveto (intof e x)(intof e y); NIL
  | ("rmoveto",dx::dy::[]) -> rmoveto (intof e dx)(intof e dy); NIL
  | ("current_point",[]) ->
      let (x,y) = current_point() in
      CV(TUPLE(Var.fresh(),INT(x)::INT(y)::[]))
  | ("lineto",x::y::[]) -> lineto (intof e x)(intof e y); NIL
  | ("rlineto",dx::dy::[]) -> rlineto (intof e dx)(intof e dy); NIL
  | ("draw_rect",x::y::w::h::[]) -> draw_rect (intof e x)(intof e y)(intof e w)(intof e h); NIL
  | ("draw_circle",x::y::r::[]) -> draw_circle (intof e x)(intof e y)(intof e r); NIL
  | ("set_line_width",i::[]) -> set_line_width (intof e i); NIL
  | ("draw_char",c::[]) -> draw_char (charof e c); NIL
  | ("draw_string",s::[]) -> draw_string (stringof e s); NIL
  | ("set_font",s::[]) -> set_font (stringof e s); NIL
  | ("set_text_size",n::[]) -> set_text_size (intof e n); NIL
  | ("text_size",s::[]) -> 
      let (i,j) = text_size(stringof e s) in
      CV(TUPLE(Var.fresh(),INT(i)::INT(j)::[]))
  | ("fill_rect",x::y::w::h::[]) -> fill_rect (intof e x)(intof e y)(intof e w)(intof e h); NIL
  | ("fill_circle",x::y::r::[]) -> fill_circle (intof e x)(intof e y)(intof e r); NIL
  | ("wait_next_event",el::[]) ->
      let getevent v = 
	match (fst (variantof e v)) with
	  "Button_down" -> Button_down
      	| "Button_up" -> Button_up
      	| "Key_pressed" -> Key_pressed
      	| "Mouse_motion" -> Mouse_motion
      	| "Poll" -> Poll
        | _ -> raise (Failure ("Function graphics.wait_next_event: event not defined"))
      in
      let retStat s = 
	CV(TUPLE(Var.fresh(),INT(s.mouse_x)
		 ::INT(s.mouse_y)::BOOL(s.button)
		 ::BOOL(s.keypressed)::CHAR(s.key)::[]
		))
      in
      let el = List.map getevent (listof e el) in
      retStat (wait_next_event el)
  | ("mouse_pos",[]) -> 
      let (i,j) = mouse_pos() in
      CV(TUPLE(Var.fresh(),INT(i)::INT(j)::[]))
  | ("button_down",[]) -> AV(BOOL(button_down()))
  | ("read_key",[]) -> AV(CHAR(read_key()))
  | ("key_pressed",[]) -> AV(BOOL(key_pressed()))
  | ("echo_keys",[]) ->
      let max = 20 in 
      let rec read_str str =
	let rec read_chars s n = 
	  let key = read_key () in
	  if (Char.code key) <> 13  (* Enter *)
	  then 
	    let _ = Bytes.set s n key in
     	    draw_char key;
      	    if n < max-1 then 
	      read_chars s (n+1)
      	    else
      	      (false,s) 
	  else 
	    (true,(Bytes.sub s 0 n))
	in
	let (ok,s) = read_chars (Bytes.create max) 0 in
	let str = String.concat "" (str::(Bytes.to_string s)::[]) in
	if ok then str else (read_str str)
      in
      CV(STRING(Var.fresh(),read_str (Bytes.to_string (Bytes.create 0))))
  | ("sound",f::d::[]) -> sound (intof e f)(intof e d); NIL
  | ("auto_synchronize",b::[]) -> auto_synchronize (boolof e b); NIL
  | ("synchronize",[]) -> synchronize(); NIL
  | ("display_mode",b::[]) -> display_mode (boolof e b); NIL
  | ("remember_mode",b::[]) -> remember_mode (boolof e b); NIL
  | (s,_) -> raise (Failure ("Function graphics."^s^" not defined or incorrect number of arguments"))

in 
let result = 
  match ci.Ccode.code with
    "Bool"::f::t     -> libBool (f,args)
  | "Char"::f::t     -> libChar (f,args)
  | "Error"::f::t    -> libError (f,args)
  | "Int"::f::t      -> libInt (f,args)
  | "List"::f::t     -> libList (f,args)
  | "Misc"::f::t     -> libMisc (f,args)
  | "Opt"::f::t      -> libOpt (f,args)
  | "Prim"::f::t     -> libPrim (f,args)
  | "String"::f::t   -> libString (f,args)
  | "Sys"::f::t      -> libSys (f,args)
  | "Graphics"::f::t -> libGraphics (f,args)
  | n::_ -> raise (Failure ("Library "^n^" not recognised"))
  | [] -> raise (Failure ("No function specified"))
in
  if ci.Ccode.alloc = 1 then  
    (* 
     * the function has been executed in a separate thread -- export the function result 
     *)
    let mytid = Thread.id (Thread.self ()) in
    Mutex.lock funsem; 
    let (t,_,cont) = ThrMap.find (mytid,ag) !funthreads in
    funthreads := ThrMap.add (mytid,ag) (t,Some result,cont) !funthreads;  (* remove this line if version 2? *)
    Mutex.unlock funsem;
(* version 1
    Mutex.lock agsem;
    let (_,_,r) = AgMap.find ag !agmap in
    Queue.add (snd e,CWAIT(mytid)) r;  (* add to run queue the thread guard *)
    Mutex.unlock agsem;
*)
(* version 2
    (match cont with 
      CCODE(ropt,_,_,p) -> 
   	(let (e,le) = update_env e result ropt in  (* read result *)
	Mutex.lock agsem;
    	let (h,e,r) = AgMap.find ag !agmap in
	Queue.add (le,p) r;  (* add the thread continuation to run queue *)
	Mutex.unlock agsem)
    | _ -> raise (Failure "System thread: thread continuation not well formed"));
*)
    NIL  (* ignored *)
  else
    result (* the same thread -- return the function result *)


(*****************
 *  Interpreter  *
 *****************)

let rec red_rules ag h e r = function
    
    (le, CCODE(ropt,ci,args,p)) ->  (* a system function call *)
      
      if ci.Ccode.alloc <> 1 then  
        (* 
         * execute in the current thread 
         *)
        let result = evalCcode (ag,(e,le),ropt,ci,args,(h,r,p)) in
        let (e,le) = update_env (e,le) result ropt in
	if ci.Ccode.alloc <> 10 then 
	  red_rules ag h e r (le, p)
	else 
	  (None,h,e,r)
      else  
        (* 
         * create a new thread 
         *)
       	(let t = Thread.create evalCcode (ag,(e,le),ropt,ci,args,(h,r,p)) in
	(* 
         * remember a new thread 
         *)
        let tid = (Thread.id t) in
        Mutex.lock funsem;
    	funthreads := ThrMap.add (tid,ag) (t,None,CCODE(ropt,ci,args,p)) !funthreads;
    	Mutex.unlock funsem;
        (*
         * add to run queue the thread guard
         *)
        Mutex.lock agsem;
        Queue.add (le,CWAIT(tid)) r;
        Mutex.unlock agsem;
        (Some ag,h,e,r))

  | (le, CWAIT(tid)) -> (* a guard of the system function thread *)
      (* 
       * Check status of the function thread 
       *)
      Mutex.lock funsem;
      let stat =
	try
	  (Some (ThrMap.find (tid,ag) !funthreads)) 
       	with _ -> None
      in
      (match stat with
      	Some (_,state,cont) ->
      	  (match state with
            None ->  (* the execution still in progress -- replace the guard *) 
              Mutex.unlock funsem;
              Mutex.lock agsem;
              Queue.add (le,CWAIT(tid)) r;
              Mutex.unlock agsem;
              (Some ag,h,e,r)
      	  | Some(result) ->
              (match cont with 
              	CCODE(ropt,ci,args,p) -> 
              	  let (e,le) = update_env (e,le) result ropt in  (* read the result *)
    	      	  funthreads := ThrMap.remove (tid,ag) !funthreads;
    	      	  Mutex.unlock funsem;
              	  red_rules ag h e r (le, p)  (* execute continuation process *)
              | _ -> raise (Failure "red_rules.CCODE: process abstraction not well formed")))
      |	_ -> (Mutex.unlock funsem; 
	      (Some ag,h,e,r))) (* ignore *)  
	
  | (le, ATOM(x,a,p)) ->        (* Atom value *)
(*
      let le = envAdd "red_rules.ATOM" x (evalAtom (e,le) a) le in
*)
      let le = EnvMap.add x (evalAtom (e,le) a) le in
      red_rules ag h e r (le, p)
	
  | (le, VAL(l,p)) ->           (* Complex value *)
      let (h,e,le) = proceedVal h e le l in
      red_rules ag h e r (le, p)

(* 
 * Core Pict
 *)

  | (le, PRL(p1,p2)) ->         (* Prl *)
      (*
  	red_rules ag h e r (le, p1);
  	red_rules ag h e r (le, p2)
  	*)
      Mutex.lock agsem;
      Queue.add (le, p2) r;
      Mutex.unlock agsem;
      red_rules ag h e r (le, p1)

  | (le, IF(a,p1,p2)) ->        (* If a then p1 else p2 *)
      if (getBool e le a) 
      then red_rules ag h e r (le, p1)
      else red_rules ag h e r (le, p2)
	  
  | (le, NEW(x,stat,p)) ->      (* Res *)
      let (h,e,le) = proceedNew h e le x stat in
      red_rules ag h e r (le, p)

  | (le, INPUT(a1,a2,stat)) ->
      proceedInput ag h e le r a1 a2 stat

  | (le, REP_INPUT(a1,a2)) ->
      proceedReplInput ag h e le r a1 a2

  | (le, TIMED_INPUT(a0,a1,a2,p,id)) ->
      proceedTimedInput ag h e le r a0 a1 a2 p id

  | (le, REP_TIMED_INPUT(a0,a1,a2,p,id)) ->
      proceedRepTimedInput ag h e le r a0 a1 a2 p id

  | (le, OUTPUT(a1,a2,stat)) ->
      let (h,e,_) = proceedOutput h e le r a1 a2 stat in
      (Some ag,h,e,r)

  | (le, SEND(a,l)) ->          (* AOut-R* *)
      let h = proceedSend h e le r a l in
      (Some ag,h,e,r)

  | (le, SKIP) ->               (* Nil *)
      (Some ag,h,e,r)

(* 
 * Agent primitives
 *)

  | (le, IFLOCAL(a,x,v,p,q)) -> (* local Test & Send *)
      proceedIfLocal ag h e le r a x v p q

  | (le, MIGRATE(a,opt,p)) ->
      (try 
	let _ = MigLockMap.find ag !miglockmap in
        (* migration is blocked -- postpone migration until lock released *)
	Mutex.lock agsem;
        Queue.add (le,MIGRATE(a,opt,p)) r;
        Mutex.unlock agsem;
        (Some ag,h,e,r)         
      with Not_found ->         
        (* migrate *)
	proceedMigrate ag h e le r a opt p)

  | (le, TERMINATE) ->          (* Terminate current agent *)
      proceedTerminate ag;
      (None,h,e,r)

(* 
 * Ignored 
 *)

  | (le, STRUCT(atom)) ->       (* atom : Types.c *)
      (Some ag,h,e,r)

  | (le, ICODE(varopt,call,p)) ->  (* not used; CCODE used instead *)
      (Some ag,h,e,r)


(* Reduction rules 
   ~~~~~~~~~~~~~~~ *)

and evalVal h e le = function
    TUPLE(x,l) ->
(*
      let le = envAdd "evalVal.TUPLE" x (CV (TUPLE (x,l))) le in
*)
      let le = EnvMap.add x (CV (TUPLE (x,l))) le in
      (h,e,le)
  | STRING(x,s) ->
(*
      let le = envAdd "evalVal.STRING" x (CV (STRING (x,s))) le in
*)
      let le = EnvMap.add x (CV (STRING (x,s))) le in
      (h,e,le)
  | DEF(x,inl,lopt,p) ->  (* Repl or Input abstraction *) 
      let e = EnvMap.add x (DF(le,DEF(x,inl,lopt,p))) e in
(*
      let le = envAdd "evalVal.DEF" x (DF(le,DEF(x,inl,lopt,p))) le in
*)
      let le = EnvMap.add x (DF(le,DEF(x,inl,lopt,p))) le in
      (h,e,le)
  | CREATE_AGENT(x,p) ->
      let h' = chanCreate None in
      let e' = envCreate (Some e) in
      let runq' = Queue.create() in
      let pe = envCreate (Some le) in
      let a = getAG (envFind x pe) in
      Queue.add (pe, p) runq';
      Mutex.lock agsem;
      agmap := AgMap.add a (h',e',runq') !agmap;  (* add the new agent to agent store *)
      Queue.add a agrunq;                         (* ... and to agent queue *)
      Mutex.unlock agsem;
      (h,e,le)


and proceedVal h e le l =
  let find_all_names le = function
      CREATE_AGENT(x,_) ->
(*
 	let a = fresh x in envAdd "evalVal.CREATE_AGENT" x (AG a) le
*)
 	let a = fresh x in EnvMap.add x (AG a) le
    | _ -> le
  in
  let (* find all agent names in definitions "agent a = ... and b = ..." *)
      le = List.fold_left (fun le a -> find_all_names le a) le l 
  in
  let (h,e,le) = List.fold_left (fun (h,e,le) a -> evalVal h e le a) (h,e,le) l in
  (h,e,le)


and getBool e le a =
  match evalAtom (e,le) a with
    AV(BOOL(b)) -> b
  | AV(_) | CV(_) | DF(_) | LV(_) | CH(_) | AG(_) -> raise (Failure "getBool: expected Boolean")
  | NIL -> raise (Failure "getBool: nil binding")


and proceedNew h e le x stat =  (* Res *)  (* REQUIRED/NOTREQUIRED ignored *)
  let c = fresh x in
(*
  let le = envAdd "proceedNew" x (CH c) le in
*)
  let le = EnvMap.add x (CH c) le in
  let h = chanAdd c (QWS[]) h in
  (h,e,le) 


and proceedOutput h e le r ca arg stat =
  match evalAtom (e,le) ca with
    CH(c) ->
      (let (h,ch) = safeChanFind c h in 
      match ch with
    	QWS[] | QRS[] ->             (* AOut-W *)
	  let h = chanUpdate c (QWS [OUT(ca,(evalAtom (e,le) arg),stat)]) h in 
	  (h,e,le)
	    
      | QWS(ws) ->                   (* AOut-W *)
	  let h = chanUpdate c (QWS (ws@[OUT(ca,(evalAtom (e,le) arg),stat)])) h in
	  (h,e,le)
	    
      | QRS((re,reader)::rest) ->    (* AOut-R *)
	  (match reader with
	    INP(_,abs,_) ->          (* reader *)
              begin
	    	match abs with
		  DEF(_,_,[Some par],p) ->
		    begin
		      let h = chanUpdate c (QRS rest) h in
(*
		      let re = envAdd "proceedOutput.INP" par (evalAtom (e,le) arg) re in
*)
		      let re = EnvMap.add par (evalAtom (e,le) arg) re in
		      Mutex.lock agsem;
		      Queue.add (re,p) r;
		      Mutex.unlock agsem;
		      (h,e,le)
		    end
	    	| _ -> raise (Failure "proceedOutput.INP: input abstraction not well formed")
	      end
	  | REPINP(a,abs,absc) ->    (* replicated reader *)
              begin
	    	match absc with
		  DEF(_,_,[Some par],p) ->
		    begin
		      let h = chanUpdate c (QRS(rest@[(re,REPINP(a,abs,absc))])) h in
(*
		      let re = envAdd "proceedOutput.REPINP" par (evalAtom (e,le) arg) re in
*)
		      let re = EnvMap.add par (evalAtom (e,le) arg) re in
		      Mutex.lock agsem;
		      Queue.add (re,p) r;
		      Mutex.unlock agsem;
		      (h,e,le)
		    end
	    	| _ -> raise (Failure "proceedOutput.REPINP: input abstraction not well formed")
	      end
	  | TIMER(_,abs,_,_) ->      (* timed reader *)
              begin
		match abs with
		  DEF(_,_,[Some par],p) ->
		    begin
		      let h = chanUpdate c (QRS rest) h in
(*
		      let re = envAdd "proceedOutput.TIMER" par (evalAtom (e,le) arg) re in
*)
		      let re = EnvMap.add par (evalAtom (e,le) arg) re in
		      Mutex.lock agsem;
		      Queue.add (re,p) r;
		      Mutex.unlock agsem;
		      (h,e,le)
		    end
	    	| _ -> raise (Failure "proceedOutput.TIMER: input abstraction not well formed")
	      end))
  | DF(fe, DEF(_,_,[Some par],p)) ->
      begin
(*
	let fe = envAdd "proceedOutput.DF" par (evalAtom (e,le) arg) fe in
*)
	let fe = EnvMap.add par (evalAtom (e,le) arg) fe in
	Mutex.lock agsem;
	Queue.add (fe,p) r;
	Mutex.unlock agsem;
	(h,e,le)
      end
  | _ -> raise (Failure "proceedSend: output expression not well formed")


and proceedSend h e le r ca argl =  (* AOut-R* *)
  let (fe,def) =  getDF (evalAtom (e,le) ca) in
  match def with
    DEF(x,_,parl,p) ->
      begin
        let f fe arg = function
            Some(par) -> EnvMap.add par (evalAtom (e,le) arg) fe
          | None ->  raise (Failure "proceedSend: no pattern variable to match argument")
        in
        let fe = List.fold_left2 f fe argl parl in
        let fe = EnvMap.add x (DF(le,def)) fe in
        Mutex.lock agsem;
        Queue.add (fe,p) r;
        Mutex.unlock agsem;
        h
      end
  | _ -> (prwarn "proceedSend: Expected replicated input expression";
          let (h,_,_)=(proceedOutput h e le r ca (hd argl) REPLICATED) in
          h)


and proceedInput ag h e le r ca abs stat =
  let (_,abs) = getDF (evalAtom (e,le) abs) in
  let c = getCH (evalAtom (e,le) ca) in
  let (h,ch) = safeChanFind c h in
  match ch with    
    QRS[]            (* Inp-R *)
  | QWS[] ->
      let h = chanUpdate c (QRS [(le,INP(ca,abs,stat))]) h in
      (Some ag,h,e,r)

  | QRS(cq) ->       (* Inp-R *)
      let h = chanUpdate c (QRS(cq@[(le,INP(ca,abs,stat))])) h in
      (Some ag,h,e,r)
	
  | QWS(wr::rest) -> (* AInp-W *)
      (match wr with
	OUT(_,v,_) ->
	  begin
	    match abs with
	      DEF(_,_,[Some par],p) ->
	    	let h = chanUpdate c (QWS rest) h in
(*
		let le = envAdd "proceedInput" par v le in
*)
		let le = EnvMap.add par v le in
		red_rules ag h e r (le,p)
	    | _ -> raise (Failure "proceedInput: input abstraction not well formed")
	  end)


and proceedReplInput ag h e le r ca abs =
  let (_,absc) = getDF (evalAtom (e,le) abs) in
  let c = getCH (evalAtom (e,le) ca) in
  let (h,ch) = safeChanFind c h in
  match ch with    
    QRS[]            (* Inp-R *)
  | QWS[] ->
      let h = chanUpdate c (QRS [(le,REPINP(ca,abs,absc))]) h in
      (Some ag,h,e,r)
	
  | QRS(cq) ->       (* Inp-R *)
      let h = chanUpdate c (QRS(cq@[(le,REPINP(ca,abs,absc))])) h in
      (Some ag,h,e,r)

  | QWS(wr::rest) -> (* AInp-W *)
      (match wr with
	OUT(_,v,_) ->
	  begin
	    match absc with
	      DEF(_,_,[Some par],p) ->
	    	let h = chanUpdate c (QWS rest) h in
(*
		let lecpy = envAdd "proceedReplInput" par v le in
*)
		let lecpy = EnvMap.add par v le in
		Mutex.lock agsem;
		Queue.add (le, REP_INPUT(ca,abs)) r;
		Mutex.unlock agsem;
               	red_rules ag h e r (lecpy,p)
	    | _ -> raise (Failure "proceedReplInput: input abstraction not well formed")
	  end)


and proceedTimedInput ag h e le r timeout ca abs elsep seq =
  let (_,absc) = getDF (evalAtom (e,le) abs) in
  let c = getCH (evalAtom (e,le) ca) in
  let (h,ch) = safeChanFind c h in
  match ch with
    QRS[]
  | QWS[] ->
      if seq = 0 then  (* it's a new timed input -- start a timer! *)
	let id = freshi() in
      	let h = chanUpdate c (QRS [(le,TIMER(ca,absc,Stdlib.truncate (time()),id))]) h in
	Mutex.lock agsem;
	Queue.add (le,TIMED_INPUT(timeout,ca,abs,elsep,id)) r;
	Mutex.unlock agsem;
      	(Some ag,h,e,r)
      else  (* the queue is empty so the input has been consumed -- do nothing! *)
	(Some ag,h,e,r)

  | QRS(cq) ->
      if seq = 0 then  (* it's a new timed input -- start a timer! *)
	let id = freshi() in
      	let h = chanUpdate c (QRS(cq@[(le,TIMER(ca,absc,Stdlib.truncate (time()),id))])) h in
	Mutex.lock agsem;
	Queue.add (le,TIMED_INPUT(timeout,ca,abs,elsep,id)) r;
	Mutex.unlock agsem;
      	(Some ag,h,e,r)
      else  (* check if the input is in the queue *)
	let rec check found stat seq l res =
	  try
	    let head = (hd l) in
	    match head with
	      (oldle,TIMER(ca,abs,start,id)) -> 
		if seq = id then  (* got it! -- check time *)
		  begin  (* maybe it'd be useful to use current 'le'? *)
		    if ((Stdlib.truncate (time()))-start) > (intof (e,oldle) timeout) then
                      (* report timeout, remove the timer from the list *)
                      check true true seq (tl l) res
		    else  (* mark that exists *)
		      check true stat seq (tl l) (head::res)
		  end
		else check found stat seq (tl l) (head::res)  (* leave unchanged *)
	    | _ -> check found stat seq (tl l) (head::res) (* leave unchanged *)
	  with hd -> (found, stat, res)  (* end of list -- return a freshly built list *)
	in
	let (present,timeouted,new_cq) = check false false seq cq [] in
	let h = chanUpdate c (QRS (rev new_cq)) h in  (* reverse the queue and store in heap *)
	if present then
	  if timeouted then  (* invoke exception on timeout *)
	    red_rules ag h e r (le,elsep)
	  else begin  (* store the timed input back in the run-queue *)
	    Mutex.lock agsem;
	    Queue.add (le,TIMED_INPUT(timeout,ca,abs,elsep,seq)) r;
	    Mutex.unlock agsem;
     	    (Some ag,h,e,r)
	  end
	else  (* the input has already been consumed -- do nothing! *)
	  (Some ag,h,e,r)

  | QWS(wr::rest) ->
      if seq = 0 then  (* it's a new timed input -- match with the output! *)
	(match wr with
	  OUT(_,v,_) ->
	    begin
	      match absc with
		DEF(_,_,[Some par],p) ->
	    	  let h = chanUpdate c (QWS rest) h in
(*
		  let le = envAdd "proceedTimedInput" par v le in
*)
		  let le = EnvMap.add par v le in
                  red_rules ag h e r (le,p)  (* execute the input abstraction, ignore timeout exception *)
	      | _ -> raise (Failure "proceedTimedInput: input abstraction not well formed")
	    end)
      else  (* the input has already been consumed -- do nothing!  *)
	(Some ag,h,e,r)


and proceedRepTimedInput ag h e le r timeout ca abs elsep seq =
  let (_,absc) = getDF (evalAtom (e,le) abs) in
  let c = getCH (evalAtom (e,le) ca) in
  let (h,ch) = safeChanFind c h in
  match ch with
    QRS[]
  | QWS[] ->
      if seq = 0 then  (* it's a new timed input -- start a timer! *)
	let id = freshi() in
      	let h = chanUpdate c (QRS [(le,TIMER(ca,absc,Stdlib.truncate (time()),id))]) h in
	Mutex.lock agsem;
	Queue.add (le,REP_TIMED_INPUT(timeout,ca,abs,elsep,id)) r;
	Mutex.unlock agsem;
      	(Some ag,h,e,r)
      else  (* the queue is empty so the input has been consumed -- create a fresh copy of timed input! *)
	proceedRepTimedInput ag h e le r timeout ca abs elsep 0

  | QRS(cq) ->
      if seq = 0 then  (* it's a new timed input -- start a timer! *)
	let id = freshi() in
      	let h = chanUpdate c (QRS(cq@[(le,TIMER(ca,absc,Stdlib.truncate (time()),id))])) h in
	Mutex.lock agsem;
	Queue.add (le,REP_TIMED_INPUT(timeout,ca,abs,elsep,id)) r;
	Mutex.unlock agsem;
      	(Some ag,h,e,r)
      else  (* check if the input is in the queue *)
	let rec check found stat seq l res =
	  try
	    let head = (hd l) in
	    match head with
	      (oldle,TIMER(ca,abs,start,id)) -> 
		if seq = id then  (* got it! -- check time *)
		  begin  (* maybe it'd be useful to use current 'le'? *)
		    if ((Stdlib.truncate (time()))-start) > (intof (e,oldle) timeout) then
                      (* report timeout, remove the timer from the list *)
                      check true true seq (tl l) res
		    else  (* mark that exists *)
		      check true stat seq (tl l) (head::res)
		  end
		else check found stat seq (tl l) (head::res)  (* leave unchanged *)
	    | _ -> check found stat seq (tl l) (head::res) (* leave unchanged *)
	  with hd -> (found, stat, res) (* end of list -- return a freshly built list *)
	in
	let (present,timeouted,new_cq) = check false false seq cq [] in
	let h = chanUpdate c (QRS (rev new_cq)) h in  (* reverse the queue and store in heap *)
	if present then
	  if timeouted then  (* invoke exception on timeout *)
	    begin
	      (* keep fresh copy of timed input in the run queue *)
	      Mutex.lock agsem;
	      Queue.add (le,REP_TIMED_INPUT(timeout,ca,abs,elsep,0)) r;
	      Mutex.unlock agsem;
	      (* execute timeout exception *)
	      red_rules ag h e r (le,elsep)
	    end
	  else begin  (* store the timed input back in the run-queue *)
	    Mutex.lock agsem;
	    Queue.add (le,REP_TIMED_INPUT(timeout,ca,abs,elsep,seq)) r;
	    Mutex.unlock agsem;
     	    (Some ag,h,e,r)
	  end
	else  (* the input has already been consumed -- create a fresh copy of timed input!  *)
	  proceedRepTimedInput ag h e le r timeout ca abs elsep 0

  | QWS(wr::rest) ->
      if seq = 0 then  (* it's a new timed input -- match with the output! *)
	(match wr with
	  OUT(_,v,_) ->
	    begin
	      match absc with
		DEF(_,_,[Some par],p) ->
	    	  let h = chanUpdate c (QWS rest) h in
(*
		  let le = envAdd "proceedRepTimedInput" par v le in
*)
		  let le = EnvMap.add par v le in
                  (* keep fresh copy of timed input in the run queue *)
		  Mutex.lock agsem;
		  Queue.add (le,REP_TIMED_INPUT(timeout,ca,abs,elsep,0)) r;
		  Mutex.unlock agsem;
                  (* execute the input abstraction, ignore timeout exception *)
                  red_rules ag h e r (le,p)
	      | _ -> raise (Failure "proceedRepTimedInput: input abstraction not well formed")
	    end)
      else  (* the input has already been consumed -- create fresh copy of timed input!  *)
	proceedRepTimedInput ag h e le r timeout ca abs elsep 0


and proceedIfLocal ag h e le r a x v p q =  (* local Test & Send *)
  let a = getAG (evalAtom (e,le) a) in
  if (a = ag) then
    let (h,e,le) = proceedOutput h e le r x v (UNKNOWN) in
    red_rules ag h e r (le,p)
  else
    try
      let (h',e',r') = AgMap.find a !agmap in
      let (h',_,_) = proceedOutput h' e le r' x v (UNKNOWN) in
      Mutex.lock agsem;  
      agmap := AgMap.add a (h',e',r') !agmap;  (* update agent store *)
      (try
(*
      	let _ = Hashtbl.find agwait a in
      	Hashtbl.remove agwait a;  (* remove agent from waiting room *)
*)
      	let _ = AgWait.find a !agwait in
      	agwait := AgWait.remove a !agwait;  (* remove agent from waiting room *)
      	Queue.add a agrunq;           (* ... and put it back into agent queue *)
      with
      	Not_found -> ());
      Mutex.unlock agsem;
      red_rules ag h e r (le,p)
    with 
      Not_found -> (* agent is not here -- do else *)
      	red_rules ag h e r (le,q)


and proceedMigrate ag h e le r s opt p =
  (* 
   * Prepare for migration.
   *)
  let (host_name,port) = getSite (evalAtom (e,le) s) in
  let (official_name, host_addr) = 
    try (host_name, inet_addr_of_string host_name)
    with Failure _ ->
      try 
	let hostptr = gethostbyname host_name in
	(hostptr.h_name, hostptr.h_addr_list.(0))
      with Not_found ->
      	Error.err ("Migration failed: Unknown remote host "^host_name)
  (*    	; exit 2   to avoid warning *)
  in
  let dottedAddr = string_of_inet_addr host_addr in
  if (dottedAddr=(!my_host_addr)) && (port=(!serverPort)) then
    begin
      prlog("Migration to current host "^(gethostname())^":"^(string_of_int port)^" ignored");
      red_rules ag h e r (le,p)
    end
  else
    begin
      (*
       * Wait until all system threads of the migrating agent terminate
      *)
      let wait (tid,a) (t,state,cont) = 
    	if a = ag then 
	  Thread.join t  (* wait until thread terminates *)
    	else ()  (* skip *)
      in
      ThrMap.iter wait !funthreads; 
      (* 
       * Extract any results and continuations of thread execution
       *)
      let getresult (tid,a) (t,state,cont) l = 
    	if a = ag then 
	  (match state with
	    None -> l (* no result *)
	  | Some(result) -> 
              (match cont with 
            	CCODE(ropt,_,_,p) -> 
		  (let (e,le) = update_env (e,le) result ropt in  (* read result *)
		  Mutex.lock agsem;
		  Queue.add (le,p) r;  (* add the thread continuation to the run queue *)
		  Mutex.unlock agsem); 
		  (tid,a)::l
	      | _ -> Error.bug "Migration failed: Thread continuation not well formed"))
    	else l  (* skip *)
      in
      Mutex.lock funsem;
      let l = ThrMap.fold getresult !funthreads [] in
      let removethread m a = ThrMap.remove a m in  
      funthreads := List.fold_left removethread !funthreads l;
      (*  funthreads := ThrMap.remove (tid,ag) !funthreads; *)
      Mutex.unlock funsem;
      (*
       * Use sockets for communication
       *)
      prlog("Migration to host "^dottedAddr^":"^(string_of_int port));
      let (ok,sfd) =
        let sfd = try 
          socket PF_INET SOCK_STREAM 0
        with _ -> Error.err ("Fatal error: Can't open stream socket")
        in
	let rec try_connect i =
          (try 
(*          setsockopt sfd SO_KEEPALIVE true; 
            clear_nonblock sfd;
            setsockopt_float sfd SO_SNDTIMEO 500.0; 
            setsockopt_float sfd SO_RCVTIMEO 500.0; 
*)
            setsockopt sfd SO_REUSEADDR true;
            connect sfd (ADDR_INET(host_addr, port));
	    (true,sfd)
          with Unix_error(errcode,f,param) ->
	    if i=0 then 
	      (prwarn ("Can't connect to "^(string_of_inet_addr host_addr)^":"^(string_of_int port));
	       (* Error.err ((error_message errcode)^" when calling function "^f^" "^param); *)
	       (false,sfd))
	    else (sleep 1; try_connect (i-1))
	  )
	in
	try_connect(30)  (* timeout for connect = 30 sec *)
      in
      if ok then
 	begin
          (*
	   * Build and send agent transmissible closure.
       	   *)
	  Mutex.lock agsem;
	  let trans_closure = (ag,h,e,r,le,p) in
	  let oc = out_channel_of_descr sfd in
 	  output_value oc trans_closure;
          close_out oc; 
          (* Mutex.lock agsem; *)
	  agmap := AgMap.remove ag !agmap;  (* remove the local copy of the agent *)
	  Mutex.unlock agsem;
	  (None,h,e,r)
        end
      else
     	begin
	  prlog("Migration ignored. Can't connect to "
                             ^(string_of_inet_addr host_addr)^":"
                             ^(string_of_int port));
          if (opt = 0) then red_rules ag h e r (le,p) else (None,h,e,r)
    	end
    end


and proceedTerminate ag = 
  Mutex.lock agsem;
  try
    let _ =  AgMap.find ag !agmap in
    let kill (tid,a) (t,state,cont) = 
      if a = ag then 
	try
     	  Thread.kill t; 
	  funthreads := ThrMap.remove (tid,ag) !funthreads;
          prerr_endline "Nomadic Pict: I/O thread killed (since agent terminated)."
	with
	  _ -> 
            prerr_endline "Nomadic Pict: Kill thread."
      else ()  (* skip *)
    in
    Mutex.lock funsem; 
    ThrMap.iter kill !funthreads;  (* kill any system threads of current agents *)
    Mutex.unlock funsem;
    agmap := AgMap.remove ag !agmap;
    Mutex.unlock agsem
  with
    Not_found ->
      Mutex.unlock agsem;
      Error.bug "proceedTerminate: agent not found"


(* Scheduler
   ~~~~~~~~~ *)

(* For debugging
   print_endline "Executing agent: "; Var.format ag; 
   print_endline "The agent first process term:"; 
   Mutex.lock agsem;
   let (le,p) = Queue.take r in
   caml_formatProc (fun _ -> ()) p; 
   print_newline();
   Queue.add (le,p) r; 
   Mutex.unlock agsem;
   print_endline "These are all agents:"; 
   AgMap.iter (fun k _ -> Var.format k; print_string " ") agh;
   print_newline(); print_endline "end."; flush stdout;
*)
	
let executeAgent() =
  while true do
    let rec takeAgent() =
      try
	Mutex.lock agsem;  (* Oct-11-2002 *)
      	let ag = Queue.take agrunq in
      	let (h,e,r) = AgMap.find ag !agmap in
	Mutex.unlock agsem; (* Oct-11-2002 *)
      	(ag,h,e,r)
      with 
      	Empty ->  (* No agents in the agent queue *)
	  Mutex.unlock agsem; (* Oct-11-2002 *)
	  Mutex.lock npvm_sem;
	  npvm_suspended := true;
	  Condition.wait npvm_newagent npvm_sem;  (* wait until a new agent is received *)
	  Mutex.unlock npvm_sem; (* freshly added *)
	  takeAgent()
      | Not_found ->  (* the agent has terminated *)
	  Mutex.unlock agsem; (* Oct-11-2002 *)
      	  takeAgent()
    in
    let (ag,h,e,r) = takeAgent() in
    Mutex.lock agsem;
    try
      let proc_term = Queue.take r in
      Mutex.unlock agsem;
      let (agopt,h,e,r) = red_rules ag h e r proc_term in
      match agopt with
      	Some ag ->
    	  Mutex.lock agsem;
 	  agmap := AgMap.add ag (h,e,r) !agmap;  (* update agent store *)
	  Queue.add ag agrunq;  (* restore in the agent queue *)
	  Mutex.unlock agsem
      | None -> ()  (* the agent has terminated or migrated away *)
    with
      Empty ->
(*
      	Hashtbl.add agwait ag ();  (* put the agent in the waiting room *) 
*)
	agwait := AgWait.add ag () !agwait;
      	Mutex.unlock agsem;
      	flush Stdlib.stdout
  done

let executeAgent_main() = handle_unix_error executeAgent()


(****************
 *  I/O Server  *
 ****************)


let readClosure (new_sfd, clnt_name) =
  let (dottedAddr,port) = match clnt_name with
    ADDR_INET(host_addr,port) ->
      let dottedAddr = string_of_inet_addr host_addr in
      prlog("A new connection from "^dottedAddr^":"^(string_of_int port));
      (dottedAddr,port)
  | ADDR_UNIX _ ->
      Error.err "I/O server: Connection from UNIX domain?!"
  in
  let ic = in_channel_of_descr new_sfd in
  try 
    let (x, h,e,r,le,p) = (input_value ic : agent_t) in
    close_in ic;
    prlog("A new agent arrived, ID="^(name x));
    Mutex.lock agsem;
    Queue.add (le, p) r;                  (* Bogus! See TO DO *)
    agmap := AgMap.add x (h,e,r) !agmap;  (* add a new agent to agent store *)
    Queue.add x agrunq;                   (* add a new entry to agent queue *)
    Mutex.unlock agsem;
    if !npvm_suspended then
      (npvm_suspended := false;
       Mutex.lock npvm_sem; (* freshly added *)
       Condition.signal npvm_newagent;
       Mutex.unlock npvm_sem);
  with Unix_error(errcode,f,param) ->  (* End_of_File *)
    (prwarn ("Broken connection with "^dottedAddr^":"^(string_of_int port)); 
     close_in ic;
     Error.err ((error_message errcode)^" when calling function "^f^" "^param))
      


let nomadicServer portOpt =
  (* 
   * Open a TCP socket (an Internet stream socket).
   *)
  let sfd = 
    try socket PF_INET SOCK_STREAM 0 
    with _ -> Error.err ("I/O server: Can't open stream socket")
  in
  (*
   * Bind our local address so that the clients can send to us.
   *)
(*  setsockopt sfd SO_REUSEADDR true; *)
  (match portOpt with
  | Some port ->
      serverPort := port;
      (* bind to the port provided *)
      bind sfd (ADDR_INET(inet_addr_any, port))  (* NOTE: must be inet_addr_any, not gethostbyname()!! *)
  | None -> 
      (* bind to the first unused port *)
      let pr = ref 5000 in
      while
	try
	  bind sfd (ADDR_INET(inet_addr_any, !pr)); false
      	with Unix_error (EADDRINUSE,syscall,_) -> true
      do
        incr pr
      done;
      serverPort := !pr;
      print_endline ("Host address and port number not specified (value "^(string_of_int !pr)^" chosen)")
  );
  listen sfd 50;  (* 5 used to be a max. value *)
  (*
   * Start a Nomadic Pict Virtual Machine thread.
   *)
  (*
  let npvm_th = Thread.create executeAgent_main() in
   *)
  let _ = Thread.create executeAgent_main() in
  (*
   * Wait for a connection from a client process. 
   *)
  while true do
    let (new_sfd, clnt_name) = 
      try accept sfd
      with _ -> Error.err ("I/O server: Accept error")
    in
    (* 
     *  Communicate with the client
     *)
    readClosure (new_sfd, clnt_name)
  done


let nomadicServer_main portOpt = 
  handle_unix_error nomadicServer portOpt


(* Top-level
   ~~~~~~~~~ *)

let execute imports prog =
  let x = fresh (Var.fresh()) in (* create a name of the top agent *)
  let h = chanCreate None in     (* create a heap *)
  let e = envCreate None in      (* create an environment *)
  let r = Queue.create() in      (* create a process queue (fifo) *)
  (*
   * Add any program constants to the agent environment.
   *)
  let e = List.fold_left (fun e c -> 
    let (h,e,le) = evalVal h e e c in le) e prog.constants in
  (* 
   * Add any constants from imported modules to the agent environment.
   *)
  let f env import = List.fold_left (fun e c -> 
    let (h,e,le) = evalVal h e e c in le) env import.constants in
  let e = List.fold_left f e imports in
  (* 
   * Create and initiate a global environment.  (* See TO DO *)
   *)
  globe := e;
  Queue.add (envCreate None, prog.proc) r;  (* add the process with a null local environment *)
  agmap := AgMap.add x (h,e,r) !agmap;      (* add a new agent closure to agent store *)
  Queue.add x agrunq;                       (* add a new entry in the agent queue *)
  (* 
   * Start the I/O server thread.
   *)
  let srv_th = Thread.create nomadicServer_main !serverPortOpt in
  Thread.join srv_th; 
  exit 0

(*
open Unix

  let read_code file_name = 
  let chan_in = open_in_gen [Open_rdonly] 0 file_name in
(*  let chan_in = open_in file_name in *)
  let value = (input_value chan_in : Inter.proc) in
  close_in chan_in; 
  value

let main () =
  if Array.length Sys.argv = 2 then 
    begin
      print_endline "START!";
      let program = VAL([], NEW((Var.createVar "pict_b" 1),REQUIRED,
        read_code Sys.argv.(1)
      )) in
      let () = initAgentProcess program in
      let () = semantics() in
      print_endline "STOP!";
      exit 0
    end else begin
      Error.err "Usage: pmach <program_source_file>";
      exit 1
    end

let () = handle_unix_error main ()
*)


(*
(* ocamlc  -custom unix.cma -cclib -lunix gethost.ml  *)

open Unix

let f = 
  
(* this below returns, e.g. glia.cl.cam.ac.uk *)
  
  print_endline (gethostname());
    
(* and this below returns 128.232.8.29 *)
    
  print_endline (string_of_inet_addr 
                   (gethostbyname (gethostname())).h_addr_list.(0) 
                )
*)
