(*
 * Functions to check for well-kinded type expressions
 * (and to convert types into our internal formats).
 *)

open Types

let checkDuplicates fi l1 l2 =
  let f seen = function
    LABELED(n,t) ->
      if Misc.StringSet.mem n seen then
	Error.errfAt fi (fun () ->
	  Format.print_string "Duplicate record label: ";
          Format.print_string n
        )
      else
	Misc.StringSet.add n seen
  | ANON(_) | TY(_) ->
      seen
  in let base =
    List.fold_left (fun s ft -> match ft with
        LABELED(n,_) -> Misc.StringSet.add n s
      | ANON(_) | TY(_) -> s
    ) Misc.StringSet.empty l1
  in
    List.fold_left f base l2

let checkVartDuplicates fi l1 l2 =
  let f seen (n,t) =
      if Misc.StringSet.mem n seen then
	Error.errfAt fi (fun () ->
	  Format.print_string "Duplicate variant label: ";
          Format.print_string n
        )
      else
	Misc.StringSet.add n seen
  in let base =
    List.fold_left (fun s ft -> match ft with
        (n,_) -> Misc.StringSet.add n s
    ) Misc.StringSet.empty l1
  in
    List.fold_left f base l2

(* Convert types to a format used by the language translation. *) 
 
let findTVId fi c ec s = Enco.lookupTVId fi (Types.preLookupTag "prekinding.findTVId." fi s c) ec

let rec scanT c ec = function
    Syntax.TOP(fi) -> Enco.TOP(fi)
  | Syntax.TVAR(fi,s) -> Enco.TVAR(fi,findTVId fi c ec s)
  | Syntax.TABS(fi,l,t) -> 
      Enco.TABS(fi,List.map (fun (s,k) -> (findTVId fi c ec s,k)) l, scanT c ec t)
  | Syntax.REC(fi,s,k,t) -> Enco.REC(fi,findTVId fi c ec s, k, scanT c ec t)
  | Syntax.TAPP(fi,t,l) -> Enco.TAPP(fi, scanT c ec t, List.map (fun t -> scanT c ec t) l)
  | Syntax.RECORD(fi,l) -> Enco.RECORD(fi, List.map (scanFieldT c ec) l)
  | Syntax.WITH(fi,t,l) -> Enco.WITH(fi,scanT c ec t,List.map (scanFieldT c ec) l)
  | Syntax.WHERE(fi,t,l) -> Error.bugAt fi "WHERE types not implemented"
  | Syntax.CHAN(fi,t) -> Enco.CHAN(fi,scanT c ec t)
  | Syntax.OCHAN(fi,t) -> Enco.OCHAN(fi,scanT c ec t)
  | Syntax.ICHAN(fi,t) -> Enco.ICHAN(fi,scanT c ec t)
  | Syntax.RCHAN(fi,t) -> Enco.RCHAN(fi,scanT c ec t)
  | Syntax.STRING(fi) -> Enco.STRING(fi)
  | Syntax.BOOL(fi) -> Enco.BOOL(fi)
  | Syntax.CHAR(fi) -> Enco.CHAR(fi)
  | Syntax.INT(fi) -> Enco.INT(fi)
  | Syntax.SITE(fi) -> Enco.SITE(fi)
  | Syntax.LOC(fi) -> Enco.LOC(fi)
  | Syntax.AGENTT(fi,t) ->Enco.AGENTT(fi,scanT c ec t)
  | Syntax.AGENT(fi) ->Enco.AGENT(fi)
  | Syntax.DYN(fi) -> Enco.DYN(fi)
  | Syntax.VART(fi,l) -> Enco.VART(fi,List.map (fun (s,t) -> (s,scanT c ec t)) l)
  | Syntax.MAP(fi,t1,t2) -> Enco.MAP(fi,scanT c ec t1, scanT c ec t2)
  | Syntax.ENCODE_TYPE(fi,l) -> (match l with
      (n::[]) -> 
      	Enco.ENCODE_TYPE(fi,findTVId fi c ec n)  (* Bogus? *)
    | _ -> Error.errAt fi "Wrong arguments in ENCODE TYPE")
	
and scanFieldT c ec = function
    Syntax.ANON(fi,t) -> Enco.ANON(fi,scanT c ec t)
  | Syntax.LABELED(fi,s,t) -> Enco.LABELED(fi,s,scanT c ec t)
  | Syntax.TY(fi,s,con) -> Enco.TYP(fi,findTVId fi c ec s,scanConstr c ec con)

and scanOptT c ec = function
    None -> None
  | Some t -> Some (scanT c ec t)
	
and scanConstr c ec = function
    Syntax.LT(t) -> Enco.LT(scanT c ec t)
  | Syntax.EQ(t) -> Enco.EQ(scanT c ec t)

(* Check for well-kinded type expressions and return internal type formats *)

let rec wellKinded c ec = function
    Syntax.TVAR(fi,n) ->
      (let get tag c ec =
  	try
	  (Types.Map.find tag ec.Enco.tvid, ec)
  	with Not_found ->
	  let nid = Enco.createId n (Enco.int_of (Enco.fresh "")) in
	  let ec = Enco.ibind 
	      (Enco.abind (Enco.constr ec tag (Types.Map.find tag c.constr)) nid tag) 
	      tag nid 
	  in (nid, ec)
      in
      try
       	let tag = Misc.StringMap.find n c.types in
 	let (tid,ec) = get tag c ec in
       	let t = match Map.find tag c.constr with LT t -> t | EQ t -> t in
       	(TVAR(Types.kindOf t,n,tag), Enco.TVAR(fi,tid), ec)
      with Not_found ->
       	Error.errfAt fi (fun () ->
	  Format.print_string "Unbound type variable: "; Format.print_string n
	    ))
  | Syntax.TABS(fi,ll,s) ->
      let (l,kl,c,ec) =
      	List.fold_left (fun (l,kl,c,ec) (n,k) ->
          let tag = freshTag() in
	  let c = abind (constr c tag (LT(mkTop k))) n tag in
  	  let nid = Enco.fresh n in
	  let ec = Enco.ibind 
	      (Enco.abind (Enco.constr ec tag (LT(mkTop k))) nid tag) tag nid 
	  in
	  ((n,tag,k) :: l, k :: kl, c, ec)
	    ) ([],[],c,ec) ll
      in let (s,ss,ec) = wellKinded c ec s
      in (TABS(Kind.ARROW(List.rev kl,Types.kindOf s),List.rev l,s),
	  Enco.TABS(fi,List.map (fun (s,k) -> 
	    (Enco.lookupTVId fi (lookupTag "prekinding.TABS" fi s c) ec,k)) ll,ss),
	  ec)
  | Syntax.REC(fi,n,k,t) -> 
      let tag = freshTag() in
      let nid = Enco.fresh n in
      let ec = Enco.ibind (Enco.abind (Enco.constr ec tag (LT(mkTop k))) nid tag) tag nid in
      let (t,tt,ec) = wellKinded (abind (constr c tag (LT(mkTop k))) n tag) ec t in
      if k <> (Types.kindOf t) then
      	Error.errfAt fi (fun()->
	  Format.print_string "Ill-kinded recursive type:";
	  Format.print_space();
	  Format.print_string "declared kind of bound variable "; 
	  Format.print_string n; 
	  Format.print_string " must be equal to kind of corresponding body")
      else
      	(REC(k,n,tag,t), Enco.REC(fi,nid,k,tt), ec)
| Syntax.TAPP(fi,s,l) ->
    let (s,ss,ec) = wellKinded c ec s in let ks = Types.kindOf s in
(* old:
    let l = List.map (wellKinded c ec) l in let kl = List.map Types.kindOf l in
*)
    let (l,ll,ec) = List.fold_left (fun (tl,ttl,ec) t -> 
      let (t,tt,ec) = wellKinded c ec t in (t::tl,tt::ttl,ec)) ([],[],ec) l in 
    let l = List.rev l in
    let ll = List.rev ll in
    let kl = List.map Types.kindOf l in
    begin
      match ks with
        Kind.ARROW(ks1,ks2) ->
          if kl = ks1 then (TAPP(ks2,s,l), Enco.TAPP(fi,ss,ll), ec)
          else Error.errfAt fi (fun()->
            Format.print_string "Ill-kinded type application of type";
            Format.print_space(); printT s; Format.print_string " : ";
	    Kind.print ks; Format.print_space();
	    Format.print_string "to types ";
	    Misc.formatList "" " " "" printT l;
	    Format.print_string " : ";
	    Misc.formatList "" " " "" Kind.print kl)
      | _ -> Error.errfAt fi (fun()->
          Format.print_string "Ill-kinded type application:";
          Format.print_space(); Format.print_string "type ";
          printT s; Format.print_string " has kind "; Kind.print ks;
          Format.print_string ","; Format.print_space();
          Format.print_string "not an arrow kind, so it cannot be applied to";
          Format.print_space(); Misc.formatList "" " " "" printT l)
    end
| Syntax.TOP(fi) -> (TOP, Enco.TOP(fi), ec)
| Syntax.RECORD(fi,[]) -> (RECORD [], Enco.RECORD(fi,[]), ec)
| Syntax.RECORD(fi,l) ->
    let (_,ec,l,ll) = List.fold_left buildFieldType (c,ec,[],[]) l in
    ignore (checkDuplicates fi [] l); 
    (RECORD(List.rev l), Enco.RECORD(fi,List.rev ll), ec)
| Syntax.WITH(fi,b,l) ->
    let (b,bb,ec) = wellKinded c ec b in
    let (_,ec,l,ll) = List.fold_left buildFieldType (c,ec,[],[]) l in
    (match expose c b with
      RECORD(bl) -> ignore (checkDuplicates fi bl l); 
	(RECORD(bl @ List.rev l), Enco.WITH(fi,bb,List.rev ll), ec)
    | _ -> Error.errfAt fi (fun()->
        Format.print_string "Left-hand side of WITH expression"; 
        Format.print_space(); printT b; Format.print_space();
        Format.print_string "is not a record type"))
| Syntax.WHERE(fi,_,_) ->
    Error.bugAt fi "WHERE types not implemented"
| Syntax.RCHAN(fi,t) -> let (t,tt,ec) = hasKindType c ec t in (RCHAN(t), Enco.RCHAN(fi,tt), ec)
| Syntax.CHAN(fi,t)  -> let (t,tt,ec) = hasKindType c ec t in (CHAN(t), Enco.CHAN(fi,tt), ec)
| Syntax.OCHAN(fi,t) -> let (t,tt,ec) = hasKindType c ec t in (OCHAN(t), Enco.OCHAN(fi,tt), ec)
| Syntax.ICHAN(fi,t) -> let (t,tt,ec) = hasKindType c ec t in (ICHAN(t), Enco.ICHAN(fi,tt), ec)
| Syntax.STRING(fi) -> (STRING, Enco.STRING(fi), ec)
| Syntax.BOOL(fi) -> (BOOL, Enco.BOOL(fi), ec)
| Syntax.CHAR(fi) -> (CHAR, Enco.CHAR(fi), ec)
| Syntax.INT(fi) -> (INT, Enco.INT(fi), ec)
| Syntax.SITE(fi) -> (SITE, Enco.SITE(fi), ec)
| Syntax.LOC(fi) -> (LOC, Enco.LOC(fi), ec)
| Syntax.AGENTT(fi,t) -> let (t,tt,ec) = hasKindType c ec t in (AGENTT(t), Enco.AGENTT(fi,tt), ec)
| Syntax.AGENT(fi) -> (AGENT, Enco.AGENT(fi), ec)
| Syntax.DYN(fi) -> (DYN, Enco.DYN(fi), ec)
| Syntax.VART(fi,[]) -> (VART [], Enco.VART(fi,[]), ec)
| Syntax.VART(fi,l) ->
    let (_,ec,l,ll) = List.fold_left buildVartFieldType (c,ec,[],[]) l in
    ignore (checkVartDuplicates fi [] l); 
    (VART(List.rev l), Enco.VART(fi,List.rev ll), ec)
| Syntax.MAP(fi,t1,t2) ->
    let (t1,tt1,ec) = hasKindType c ec t1 in
    let (t2,tt2,ec) = hasKindType c ec t2 in
    (MAP(t1,t2), Enco.MAP(fi,tt1,tt2), ec)
| Syntax.ENCODE_TYPE(fi,l) -> (match l with
    (n::[]) -> (ENCODE_TYPE(l), 
		Enco.ENCODE_TYPE(fi,Enco.lookupTVId fi (Types.lookupTag "prekinding.ENCODE_TYPE" fi n c) ec),
		ec)
  | _ -> Error.errAt fi "Wrong arguments in ENCODE TYPE")

and buildFieldType (c,ec,l,ll) = function
    Syntax.ANON(fi,t) -> 
      let (t,tt,ec) = hasKindType c ec t in 
      (c, ec, ANON(t) :: l, Enco.ANON(fi,tt) :: ll)
  | Syntax.LABELED(fi,n,t) ->
      let (t,tt,ec) = hasKindType c ec t in 
      (c, ec, LABELED(n,t) :: l, Enco.LABELED(fi,n,tt) :: ll)
  | Syntax.TY(fi,n,con) ->
      let tag = freshTag() and (con,conn,ec) = wellKindedConstr c ec con in
      let c = abind (constr c tag con) n tag in
      let nid = Enco.fresh n in
      let ec = Enco.ibind (Enco.abind (Enco.constr ec tag con) nid tag) tag nid in
      (c, ec, TY(n,tag,con) :: l, Enco.TYP(fi,nid,conn) :: ll)
	
and buildVartFieldType (c,ec,l,ll) (n,t) = 
  let (t,tt,ec) = hasKindType c ec t in (c,ec,(n,t) :: l,(n,tt) :: ll)
    
and wellKindedConstr c ec = function
    Syntax.LT t -> let (t,tt,ec) = wellKinded c ec t in (LT(t), Enco.LT(tt), ec)
  | Syntax.EQ t -> let (t,tt,ec) = wellKinded c ec t in (EQ(t), Enco.LT(tt), ec)
      
and hasKindType c ec tt =
  let (t,tt',ec) = wellKinded c ec tt in
  let k = Types.kindOf t in
  if k = Kind.TYPE then (t,tt',ec) else
  Error.errfAt (Syntax.typeInfo tt) (fun () ->
    Format.print_string "Type "; printT t; Format.print_string " of kind ";
    Kind.print k; Format.print_space();
    Format.print_string "used in a context where only kind Type is allowed")
    
let mkRecord fi l = ignore (checkDuplicates fi [] l); RECORD(l)







