(*
 * Types and simple utilties (such as printing).
 *)

let debug = Flags.createFlag "TYPES"
  "*Print low-level debugging information in the types module" false

let exactT = Flags.createFlag "exactTypes"
  "*Show complete internal representation of types" false

let printKinds = Flags.createFlag "printKinds"
  "*Print the kinds of all (non-Type) types" false

let printUVarKinds = Flags.createFlag "printUVarKinds"
  "*Show the kinds of unification variables in types" false

let printTypeString = Flags.createFlag "printTypeString"
  "@Show the type converted to string" false

type tag = Var.var
let freshTag = Var.fresh
let printTag = Var.format

let str_of a = Var.str_of a  (* Return string of tag *)
let int_of a = Var.int_of a  (* Return integer of tag *)
let name a = Var.name a      (* Return tag converted to string *)

module Map = Var.Map

type t =
  TOP
| UVAR of t option ref * tag
| TVAR of Kind.kind * string * tag
| TABS of Kind.kind * (string * tag * Kind.kind) list * t
| REC of Kind.kind * string * tag * t
| TAPP of Kind.kind * t * t list
| RECORD of field list  
| RCHAN of t
| OCHAN of t
| ICHAN of t
| CHAN of t
| STRING
| CHAR
| BOOL
| INT
| SITE
| LOC
| AGENTT of t
| AGENT
| DYN
| VART of (string * t) list
| MAP of t * t
| ENCODE_TYPE of string list
| IGNORE

and field =
  ANON of t
| LABELED of string * t
| TY of string * tag * constr

and constr =
  LT of t
| EQ of t

type 'a c =
  {types: tag Misc.StringMap.t;     (* Maps type variables to unique tags *)
   constr: constr Map.t;            (* Maps tags to constraints *)  
   vars: (t * 'a) Misc.StringMap.t} (* Maps identifiers to types, plus an 'a *)

let mkUnit = RECORD []

let rec mkTop k = match k with
  Kind.TYPE -> TOP
| Kind.ARROW(l,kb) ->
  TABS(k,List.map (fun k -> (":t",freshTag(),k)) l,mkTop kb)

(*
 * Creation functions for tagged things
 *)

let mkUvar () = UVAR(ref None,Var.fresh())

let empty =
  {types=Misc.StringMap.empty; constr=Map.empty; vars=Misc.StringMap.empty}

let abind c n tag =
  {types=Misc.StringMap.add n tag c.types; constr=c.constr; vars=c.vars}

let constr c tag constr =
  {types=c.types; constr=Map.add tag constr c.constr; vars=c.vars}

let xbind n t x c =
  {types=c.types; constr=c.constr; vars=Misc.StringMap.add n (t,x) c.vars}

let union c1 c2 =
  {types=Misc.StringMap.myunion c1.types c2.types;
   constr=Map.myunion c1.constr c2.constr;
   vars=Misc.StringMap.myunion c1.vars c2.vars}

(*** Find the Kind of a type. ***)

let kindOf = function
  TVAR(k,_,_) -> k
| TABS(k,_,_) -> k
| TAPP(k,_,_) -> k
| REC(k,_,_,_) -> k
| UVAR(_) | TOP | RECORD(_) | RCHAN(_)
| OCHAN(_) | ICHAN(_) | CHAN(_) | STRING
| CHAR | BOOL | INT -> Kind.TYPE
| SITE | LOC | AGENTT(_) | AGENT | DYN | VART(_) | MAP(_,_) -> Kind.TYPE
| ENCODE_TYPE(_) | IGNORE -> Kind.TYPE

let kindOfConstr = function
  EQ t -> kindOf t
| LT t -> kindOf t

(*** Simple formatting utilities. ***)

let openParen () = Format.print_string "("; Format.open_hvbox 2
let closeParen () = Format.print_string ")"; Format.close_box()

(*** Pretty print a type. ***)

let ppTag tag =
  Format.print_string "("; Var.format tag; Format.print_string ")"

let ppK = function
  Kind.TYPE -> ()
| k -> Format.print_string ":"; Kind.print k

let ppLevel i =
  Format.print_string "+"; Format.print_int i

let ppKindedArg (n,tag,k) =
  Format.print_string n; if !exactT then ppTag tag; ppK k

let rec ppT = function
  TVAR(_,n,tag) ->
  Format.print_string n; if !exactT then ppTag tag
| UVAR({contents = None},tag) ->
  Format.print_string "Any"; ppTag tag
| UVAR({contents = Some t},tag) ->
  if !exactT then
    (Format.print_string "Any"; ppTag tag; Format.print_string "= "; ppT t)
  else
    ppT t
| TABS(_,l,t) ->
    Misc.formatList "\(" ")=" " " ppKindedArg l; Format.print_space(); ppT t
| REC(k,n,tag,t) ->
    Format.print_string "(rec ";
    Format.print_string n; if !exactT then ppTag tag; ppK k;
    Format.print_string "="; ppT t; Format.print_string ")"
| TAPP(_,t,ta) ->
    openParen(); ppT t; Format.print_space();
    Misc.formatList "" "" " " ppT ta; closeParen()
| TOP ->
    Format.print_string "Top"
| RECORD(l) ->
    Misc.formatList "[" "]" " " ppF l
| RCHAN(t) -> Format.print_string "/"; ppT t
| CHAN(t) -> Format.print_string "^"; ppT t
| OCHAN(t) -> Format.print_string "!"; ppT t
| ICHAN(t) -> Format.print_string "?"; ppT t
| STRING -> Format.print_string "String"
| CHAR -> Format.print_string "Char"
| BOOL -> Format.print_string "Bool"
| INT -> Format.print_string "Int"
| SITE -> Format.print_string "Site"
| LOC -> Format.print_string "Loc"
| AGENTT(t) -> Format.print_string "AgentT("; ppT t; Format.print_char ')'
| AGENT -> Format.print_string "Agent"
| DYN -> Format.print_string "Dyn"
| VART(l) -> 
    Misc.formatList "{" "}" " " ppVF l
| MAP(t1,t2) -> Format.print_string "Map "; ppT t1; Format.print_string " "; ppT t2 
| ENCODE_TYPE(l) -> Format.print_char '{'; 
    Misc.formatList "[" "]" " " Format.print_string l;
    Format.print_char '}'; 
| IGNORE -> Format.print_string "Ignore"

and ppF = function
  ANON(t) -> ppT t
| LABELED(n,t) -> Format.print_string n; Format.print_string " = "; ppT t
| TY(n,tag,con) ->
  Format.print_string "#"; Format.print_string n;
  if !exactT then ppTag tag; ppCon con

and ppVF (n,t) =
  Format.print_string n ; Format.print_string ">"; ppT t

and ppCon = function
  LT(TOP) -> ()
| LT(t) -> Format.print_string "<"; ppT t
| EQ(t) -> Format.print_string "="; ppT t

let printConstr = function
  LT(t) -> Format.print_string "<"; ppT t
| EQ(t) -> Format.print_string "="; ppT t

let printT t =
  let k = kindOf t in
  if !printKinds && k <> Kind.TYPE then
    (Format.print_string " : "; Kind.print k);
  Format.open_hvbox 2; ppT t; Format.close_box()

let displayT s = Format.print_break 1 3; printT s

let formatVarEntry n (t,_) =
  Format.open_hvbox 2; Format.print_string n; Format.print_string " : ";
  Format.print_cut(); printT t; Format.close_box();
  Format.print_newline()

let formatTyEntry constrs n tag =
  Format.open_hvbox 2; Format.print_string n;
  if !exactT then ppTag tag; ppCon (Map.find tag constrs); Format.close_box();
  Format.print_newline()

let printC c =
  Format.open_hovbox 0;
    Misc.StringMap.iter (formatTyEntry c.constr) c.types;
    Misc.StringMap.iter formatVarEntry c.vars;
  Format.close_box()

(*** Iterate over all value bindings in a context. ***)

let mapBindings f c =
  {types=c.types; constr=c.constr;
(* ocaml-1.05: vars=Misc.StringMap.map (fun _ x -> f x) c.vars} *)
   vars=Misc.StringMap.map f c.vars}
let allBindings f c =
  Misc.StringMap.iter (fun _ x -> f x) c.vars

let lookupBound tag c =
  try
    match Map.find tag c.constr with LT t -> t | EQ _ -> raise Not_found
  with Not_found ->
    Error.bugf (fun ()->
      Format.print_string "lookupBound: tag: "; Var.format tag
    )

let lookupAbbrv tag c =
  try
    match Map.find tag c.constr with LT _ -> None | EQ t -> Some t
  with Not_found ->
    Error.bugf (fun ()->
      Format.print_string "lookupAbbrv: tag: "; Var.format tag
    )

let lookupTag str fi n c =
  try
    Misc.StringMap.find n c.types
  with Not_found ->
    Error.errfAt fi (fun () ->
      Format.print_string (str^"Types.LookupTag: Unbound type variable: "); Format.print_string n)
      
let preLookupTag str fi n c =
  try
    Misc.StringMap.find n c.types
  with Not_found ->
    Var.fresh()

      
(*
 * Substitution
 *)

let rec sub env = function
  TOP -> TOP
| TVAR(_,_,tag) as s -> (try sub env (Map.find tag env) with Not_found -> s)
| UVAR({contents = None},_) as s -> s
| UVAR({contents = Some s},_) -> sub env s
| TABS(k,l,s) ->
    let (l,env) =
      List.fold_left (fun (l,env) (n,tag,k) ->
	  let newTag = Var.fresh() in let v = TVAR(k,n,newTag) in
        ((n,newTag,k) :: l, Map.add tag v env)
	) ([],env) l
    in
	TABS(k,List.rev l,sub env s)
| REC(k,n,tag,s) -> 
    let newTag = Var.fresh() in let v = TVAR(k,n,newTag) in
    REC(k,n,newTag,sub (Map.add tag v env) s)
| TAPP(k,t,a) -> TAPP(k,sub env t,List.map (sub env) a)
| RECORD(l) -> RECORD(List.rev(snd(List.fold_left subField (env,[]) l)))
| RCHAN(t) -> RCHAN(sub env t)
| OCHAN(t) -> OCHAN(sub env t)
| ICHAN(t) -> ICHAN(sub env t)
| CHAN(t) -> CHAN(sub env t)
| STRING -> STRING
| CHAR -> CHAR
| BOOL -> BOOL
| INT -> INT
| SITE -> SITE
| LOC -> LOC
| AGENTT(t) -> AGENTT(sub env t)
| AGENT -> AGENT
| DYN -> DYN
| VART(l) -> VART(List.rev(snd(List.fold_left subVartField (env,[]) l)))
| MAP(t1,t2) -> MAP(sub env t1,sub env t2) 
| ENCODE_TYPE(l) -> ENCODE_TYPE(l)
| IGNORE -> IGNORE

and subField (env,l) = function
  ANON(t) -> (env,ANON(sub env t) :: l)
| LABELED(n,t) -> (env,LABELED(n,sub env t) :: l)
| TY(n,tag,LT t) -> 
    let f = Var.fresh() in let tv = TVAR(kindOf t,n,f) in
    (Map.add tag tv env,TY(n,f,LT(sub env t)) :: l)
| TY(n,tag,EQ t) -> 
    let f = Var.fresh() in let tv = TVAR(kindOf t,n,f) in
    (Map.add tag tv env,TY(n,f,EQ(sub env t)) :: l)

(*
and subVartField (env,l) (n,t) =
  let f = Var.fresh() in let tv = TVAR(kindOf t,n,f) in
  (Map.add tag tv env,(n,sub env t)::l)
*)

and subVartField (env,l) (n,t) = (env,(n,sub env t)::l)

let subst s tag t =
  let env = Map.add tag t Map.empty in
  if !debug then
    Debug.wrap "subst"
    (fun () -> sub env s)
    (fun () ->
       Format.print_string "type: "; printT s;
       Format.print_space();
       Format.print_string "tag: "; Var.format tag;
       Format.print_space();
       Format.print_string "target: "; printT t)
    (fun res -> printT res)
  else
    sub env s

let clone s =
  if !debug then
    Debug.wrap "clone" (fun () -> sub Map.empty s)
    (fun () -> printT s) (fun res -> printT res)
  else
    sub Map.empty s

(*
 * Exposes the "head structure" of a type expression.
 *)

let expose c s =
  let rec exp = function
    UVAR({contents = None},_) as s -> s
  | UVAR({contents = Some t},_) -> exp t
  | TVAR(_,_,tag) as s ->
    (match lookupAbbrv tag c with None -> s | Some t -> exp (clone t))    
  | TAPP(k,s,a) ->
    begin
      match exp s with
        TABS(_,l,s) ->
	  exp(List.fold_left2 (fun s (_,tag,_) t -> subst s tag t) s l a)
      | nf -> TAPP(k,nf,a)
    end
  | s -> s
in
  if !debug then
    Debug.wrap "expose"
    (fun () -> exp s)
    (fun () -> printT s)
    (fun s -> printT s)
  else
    exp s

(*
 * Promote a exposed expression to its least (proper) upper bound.
 *)

let promote c t = 
  let rec prom = function
    TVAR(k,n,i) ->
    (match lookupBound i c with TOP -> None | b -> Some(b))
  | TAPP(k,s,a) ->
    begin
      match prom s with
	None ->
        begin
	  match s with
	    TABS(_,l,s) ->
	    Some(List.fold_left2 (fun s (_,tag,_) t -> subst s tag t) s l a)
	  | _ -> None
	end
      | Some(s') -> Some(TAPP(k,s',a))
    end
  | UVAR({contents = None},_) -> None
  | UVAR({contents = Some t},_) -> prom t
  | _ -> None
in
  if !debug then
    Debug.wrap "promote"
    (fun () -> prom t)
    (fun () -> printT t)
    (function None -> Format.print_string "NO" | Some(t) -> printT t)
  else
    prom t

(*
 * Contexts
 *)

(* DEFINITION: A type expression is CONCEALED if it has any of the
   following forms
      - a beta-redex
      - a known unification variable
      - an abbreviation
      - an application whose function part is concealed
      - a (... with ...) expression
   A type expression is EXPOSED if it is not concealed.

   The PROMOTION of a type variable is its bound, unless its
   bound is top. The promotion of an application APP(a,b) is
   APP(promote(a),b) if a is a type variable.

   A type s is BASIC if it cannot be promoted.
 
   The BASIS of a type s is the smallest basic supertype of s
   (possibly s itself).
*)

(* Find the basis of s.
   (N.b.: an earlier version calculated a slightly more refined notion
    of basis: if the upper bound of A is Top, then basis(A) = A.)
*)

let rec basis c s =
  if !debug then
    Debug.wrap "basis"
    (fun () ->
       match promote c s with None -> s | Some t -> basis c (expose c t))
    (fun () -> printT s)
    (fun s -> printT s)
  else
    match promote c s with None -> s | Some t -> basis c (expose c t)

(* 
 * Destructors and checkers
 *)

let unroll fi c s =
  let rec str = function
    REC(k,n,tag,t) as r -> subst t tag r
  | TAPP(k,s1,s2) -> TAPP(k, str s1, s2)
  | _ -> Error.errfAt fi (fun() ->
    Format.print_string "Expected recursive type at head but found";
    Format.print_string " "; printT s)
in
  if !debug then
    Debug.wrap "unroll"
    (fun () -> str (basis c (expose c s)))
    (fun () -> printT s)
    (fun s -> printT s)
  else
    str (basis c (expose c s))


(*
 * [getFields fi c t] checks that [t] is a record type, returning
 * the fields of the record, or failing otherwise.
 *)
let getFields fi c t =
  match basis c (expose c t) with
    RECORD l -> l
  | IGNORE -> Error.warningfAt fi (fun () ->
      Format.print_string "Expected a record type "(*but found ; displayT t*));
      (ANON(IGNORE)::[])  (* Bogus *)
  | _ -> Error.errfAt fi (fun () ->
     Format.print_string "Expected a record type but found "; displayT t)

let getIChan fi c t =
  match basis c (expose c t) with
    CHAN t -> t
  | ICHAN t -> t
  | _ -> Error.errfAt fi (fun () ->
     Format.print_string "Expected an input-channel type but found ";
     displayT t)
  
let getOChan fi c t =
  match basis c (expose c t) with
    CHAN t -> t
  | OCHAN t -> t
  | RCHAN t -> t
  | _ -> Error.errfAt fi (fun () ->
     Format.print_string "Expected an output-channel type but found ";
     displayT t)

let getRChan fi c t =
  match basis c (expose c t) with
    RCHAN t -> t
  | _ -> Error.errfAt fi (fun () ->
     Format.print_string "Expected an responsive-channel type but found ";
     displayT t)

let getIAgent fi c t =
  match basis c (expose c t) with
    AGENTT t -> t
  | _ -> Error.errfAt fi (fun () ->
     Format.print_string "Expected an input-agent type but found ";
     displayT t)

let getOAgent fi c t =
  match basis c (expose c t) with
    AGENTT t -> t
  | _ -> Error.errfAt fi (fun () ->
     Format.print_string "Expected an output-agent type but found ";
     displayT t)

let isChan c t =
  match expose c t with CHAN _ -> true | _ -> false

let leqRChan c t =
  match basis c (expose c t) with RCHAN _ -> true | _ -> false

let isAgent c t =
  match expose c t with 
    AGENTT _ -> true
  | AGENT -> true
  | _ -> false

let typeOfX fi n c =
  try Misc.StringMap.find n c.vars with Not_found ->
  Error.errfAt fi (fun () ->
    Format.print_string "Unbound name: "; Format.print_string n
  )

let isBoundVar c n =
  Misc.StringMap.mem n c.vars

let isBoundTyVar c n =
  Misc.StringMap.mem n c.types

let fieldType fi c t n =
  match basis c (expose c t) with
    RECORD(l) ->
      (let rec loop i = function 
         ANON(_) :: rest -> loop (i+1) rest
       | LABELED(fn,ft) :: rest -> if n = fn then (ft,i) else loop (i+1) rest
       | TY(_) :: rest -> loop i rest
       | [] -> 
          Error.errfAt fi (fun () ->
            Format.print_string "label "; Format.print_string n;
	    Format.print_string " missing from type ";
            displayT t)
       in loop 0 l)
  | _ -> 
    Error.errfAt fi (fun () -> 
      Format.print_string "expected a record type but found:"; displayT t
    )

(*
 * Printing a type (O'Caml format). 
 *)

let caml_formatTag tag = Var.format tag
    
let caml_formatStr s =
  Format.print_string "\""; 
  Format.print_string s; 
  Format.print_string "\""

let rec caml_formatConstr = function
    LT(t) -> Format.print_string "LT("; caml_formatT t; Format.print_char ')'
  | EQ(t) -> Format.print_string "EQ("; caml_formatT t; Format.print_char ')'

and caml_formatT = function
    TOP -> Format.print_string "TOP"
  | UVAR({contents = None},tag) -> 
      Format.print_string "UVAR({contents = None},";
      caml_formatTag tag; Format.print_char ')'
  | UVAR({contents = Some t},tag) -> 
      Format.print_string "UVAR({contents = Some ";
      caml_formatT t; Format.print_string "},";
      caml_formatTag tag; Format.print_char ')'
  | TVAR(k,s,tag) -> 
      Format.print_string "TVAR("; 
      Kind.caml_format k; Format.print_char ','; 
      caml_formatStr s; Format.print_char ',';
      caml_formatTag tag; Format.print_char ')'
  | TABS(k,l,t) -> 
      let formatElem (s,tag,k) =
	Format.print_char '(';
  	caml_formatStr s; Format.print_char ','; 
	caml_formatTag tag; Format.print_char ',';
	Kind.caml_format k; Format.print_char ')'
      in
      Format.print_string "TABS(";
      Kind.caml_format k; Format.print_char ','; 
      Misc.formatList " [" "]" ";" formatElem l; 
      Format.print_char ','; caml_formatT t; Format.print_char ')'
  | REC(k,s,tag,t) ->
      Format.print_string "REC(";
      Kind.caml_format k; Format.print_char ','; 
      caml_formatStr s; Format.print_char ','; 
      caml_formatTag tag; Format.print_char ','; 
      caml_formatT t; Format.print_char ')'
  | TAPP(k,t,tl) ->
      Format.print_string "TAPP(";
      Kind.caml_format k; Format.print_char ','; 
      caml_formatT t; Format.print_char ','; 
      Misc.formatList " [" "]" ";" caml_formatT tl; 
      Format.print_char ')'
  | RECORD(l) ->
      let formatElem = function
	  ANON(t) ->
	    Format.print_string "ANON("; caml_formatT t; Format.print_char ')'
  	| LABELED(s,t) -> 
	    Format.print_string "LABELED(";
	    caml_formatStr s; Format.print_char ','; 
	    caml_formatT t; Format.print_char ')'
  	| TY(s,tag,constr) -> 
	    Format.print_string "TY(";
	    caml_formatStr s; Format.print_char ','; 
	    caml_formatTag tag; Format.print_char ','; 
	    caml_formatConstr constr; Format.print_char ')'
      in
      Format.print_string "RECORD(";
      Misc.formatList " [" "]" ";" formatElem l;
      Format.print_char ')'
  | RCHAN(t) -> 
      Format.print_string "RCHAN("; caml_formatT t; Format.print_char ')'
  | OCHAN(t) -> 
      Format.print_string "OCHAN("; caml_formatT t; Format.print_char ')'
  | ICHAN(t) -> 
      Format.print_string "ICHAN("; caml_formatT t; Format.print_char ')'
  | CHAN(t) -> 
      Format.print_string "CHAN("; caml_formatT t; Format.print_char ')'
  | STRING -> Format.print_string "STRING"
  | CHAR -> Format.print_string "CHAR"
  | BOOL -> Format.print_string "BOOL"
  | INT -> Format.print_string "INT"
  | SITE -> Format.print_string "SITE"
  | LOC -> Format.print_string "LOC"
  | AGENTT(t) -> Format.print_string "AGENTT("; caml_formatT t; Format.print_char ')'
  | AGENT -> Format.print_string "AGENT"
  | DYN -> Format.print_string "DYN"
  | VART(l) ->
      let formatElem (s,t) =
	Format.print_char '(';
  	caml_formatStr s; Format.print_char ','; 
	caml_formatT t; Format.print_char ')'
      in
      Format.print_string "VART(";
      Misc.formatList " [" "]" ";" formatElem l;
      Format.print_char ')'
  | MAP(t,ty) ->
      Format.print_string "MAP(";
      caml_formatT t; Format.print_char ','; 
      caml_formatT ty; Format.print_char ')'
  | ENCODE_TYPE(l) -> Format.print_string "ENCODE_TYPE("; 
      Misc.formatList " [" "]" ";" caml_formatStr l; 
      Format.print_char ')'
  | IGNORE -> Format.print_string "IGNORE"

(*
 *  Print the structure of type 'a c (O'Caml format)
 *)

let formatExactVarEntry fa n (t,a) =
  Format.open_hvbox 2; Format.print_string n; Format.print_string " : ";
  Format.print_cut(); printT t; fa a; Format.close_box();
  Format.print_newline()

let caml_formatC c fa =
  let cfTypes n tag =
    Format.print_string n; Format.print_string "->"; Var.format tag;
    Format.print_newline ();
  in
  let cfConstr tag constr =
    Var.format tag; Format.print_string "->"; caml_formatConstr constr;
    Format.print_newline ();
  in
  let cfVars n (t,a) =
    Format.print_string n; Format.print_string "-> ("; 
    caml_formatT t; Format.print_char ','; fa a; Format.print_char ')';
    Format.print_newline ();
  in
  Format.open_hovbox 0;
  Misc.StringMap.iter cfTypes c.types;
  Map.iter cfConstr c.constr;
  Misc.StringMap.iter cfVars c.vars;
  Format.close_box()

(*
 *  Convert a type to a unique string (for dynamic values)
 *)
			
(* Bogus: not sure if the convertion is correct. *)

let csTag tag = ""

let csK = function
    Kind.TYPE -> "()"
  | k -> ":"^(Kind.toString k)

let csLevel i = "+"^(string_of_int i)

let csKindedArg (n,tag,k) = (csTag tag)^(csK k)

let rec csT = function
    TVAR(_,n,tag) -> (csTag tag)
  | UVAR({contents = None},tag) -> "Any"^(csTag tag)
  | UVAR({contents = Some t},tag) -> "Any"^(csTag tag)^"= "^(csT t)
  | TABS(_,l,t) -> (Misc.listToString "\(" ")=" " " csKindedArg l)^(csT t)
  | REC(k,n,tag,t) -> "(rec "^(csTag tag)^(csK k)^"="^(csT t)^")"
  | TAPP(_,t,ta) -> "("^(csT t)^(Misc.listToString "" "" " " csT ta)^")"
  | TOP -> "Top"
  | RECORD(l) -> Misc.listToString "[" "]" " " csF l
  | RCHAN(t) -> "/"^(csT t)
  | CHAN(t) -> "^"^(csT t)
  | OCHAN(t) -> "!"^(csT t)
  | ICHAN(t) -> "?"^(csT t)
  | STRING -> "String"
  | CHAR -> "Char"
  | BOOL -> "Bool"
  | INT -> "Int"
  | SITE -> "Site"
  | LOC -> "Loc"
  | AGENTT(t) -> "AgentT "^(csT t)
  | AGENT -> "Agent"
  | DYN -> "Dyn"
  | VART(l) -> Misc.listToString "{" "}" " " csVF l
  | MAP(t1,t2) -> "Map "^(csT t1)^" "^(csT t2) 
  | ENCODE_TYPE(l) -> "Encode type ["^(Misc.listToString "" "" " " 
      (fun s -> s) l)^"]"
  | IGNORE -> "Ignore"

and csF = function
    ANON(t) -> csT t
  | LABELED(n,t) -> n^" = "^(csT t)
  | TY(n,tag,con) -> "#"^(csTag tag)^(csCon con)

and csVF (n,t) = ">"^(csT t)

and csCon = function
    LT(TOP) -> "()"
  | LT(t) -> "<"^(csT t)
  | EQ(t) -> "="^(csT t)

let toString t =
(* For debugging type Dyn:
   printT t; Format.print_newline(); 
*)
  let s = csT t
  in if !printTypeString then (Format.print_string s; s) else s


